import numpy as np
import pyqtgraph as pg

from orangewidget.tests.base import GuiTest
from orangewidget.utils.matplotlib_export import (
    scatterplot_code, numpy_repr, compress_if_all_same, numpy_repr_int
)


def add_intro(a):
    r = "import matplotlib.pyplot as plt\n" + \
        "from numpy import array\n" + \
        "plt.clf()"
    return r + a


class TestScatterPlot(GuiTest):
    def test_scatterplot_simple(self):
        plotWidget = pg.PlotWidget(background="w")
        scatterplot = pg.ScatterPlotItem()
        scatterplot.setData(
            x=np.array([1., 2, 3]),
            y=np.array([3., 2, 1]),
            size=np.array([1., 1, 1])
        )
        plotWidget.addItem(scatterplot)
        code = scatterplot_code(scatterplot)
        self.assertIn("plt.scatter", code)
        exec(add_intro(code), {})

    def test_utils(self):
        a = np.array([1.5, 2.5])
        self.assertIn("1.5, 2.5", numpy_repr(a))
        a = np.array([1, 1])
        v = compress_if_all_same(a)
        self.assertEqual(v, 1)
        self.assertEqual(repr(v), "1")
        self.assertIs(type(v), int)
        a = np.array([1, 2], dtype=int)
        v = numpy_repr_int(a)
        self.assertIn("1, 2", v)



import enum
import inspect
import pkgutil
import sys
import warnings
from operator import attrgetter
from typing import Union, Iterator, Optional

from AnyQt.QtCore import QObject, QRect, QSize, QPoint, QTextBoundaryFinder
from AnyQt.QtGui import QIcon

from orangecanvas.gui.svgiconengine import StyledSvgIconEngine


def progress_bar_milestones(count, iterations=100):
    return set([int(i*count/float(iterations)) for i in range(iterations)])


_NOTSET = object()


def deepgetattr(obj, attr, default=_NOTSET):
    """Works exactly like getattr(), except that attr can be a nested attribute
    (e.g. "attr1.attr2.attr3").
    """
    try:
        return attrgetter(attr)(obj)
    except AttributeError:
        if default is _NOTSET:
            raise
        return default


def getdeepattr(obj, attr, *arg, **kwarg):
    if isinstance(obj, dict):
        return obj.get(attr)
    return deepgetattr(obj, attr, *arg, **kwarg)


def to_html(str):
    return str.replace("<=", "&#8804;").replace(">=", "&#8805;").\
        replace("<", "&#60;").replace(">", "&#62;").replace("=\\=", "&#8800;")

getHtmlCompatibleString = to_html


def dumpObjectTree(obj, _indent=0):
    """
    Dumps Qt QObject tree. Aids in debugging internals.
    See also: QObject.dumpObjectTree()
    """
    assert isinstance(obj, QObject)
    print('{indent}{type} "{name}"'.format(indent=' ' * (_indent * 4),
                                           type=type(obj).__name__,
                                           name=obj.objectName()),
          file=sys.stderr)
    for child in obj.children():
        dumpObjectTree(child, _indent + 1)


def getmembers(obj, predicate=None):
    """Return all the members of an object in a list of (name, value) pairs sorted by name.

    Behaves like inspect.getmembers. If a type object is passed as a predicate,
    only members of that type are returned.
    """

    if isinstance(predicate, type):
        def mypredicate(x):
            return isinstance(x, predicate)
    else:
        mypredicate = predicate
    return inspect.getmembers(obj, mypredicate)


class DeprecatedSignal:
    def __init__(self, actual_signal, *args,
                 warning_text='Deprecated', emit_callback=None, **kwargs):
        self.signal = actual_signal
        self.warning_text = warning_text
        self.emit_callback = emit_callback

    def emit(self, *args, **kwargs):
        warnings.warn(
            self.warning_text,
            DeprecationWarning, stacklevel=2
        )
        if self.emit_callback:
            self.emit_callback(*args, **kwargs)
        return self.signal.emit(*args, **kwargs)

    def __getattr__(self, item):
        return self.__signal.item


def enum_as_int(value: Union[int, enum.Enum]) -> int:
    """
    Return a `enum.Enum` value as an `int.

    This is function intended for extracting underlying Qt5/6 enum
    values specifically with PyQt6 where most Qt enums are represented
    with `enum.Enum` and lose their numerical value.

    >>> from PyQt6.QtCore import Qt
    >>> enum_as_int(Qt.Alignment.AlignLeft)
    1
    """
    if isinstance(value, enum.Enum):
        return int(value.value)
    else:
        return int(value)


def dropdown_popup_geometry(
        size: QSize, origin: QRect, screen: QRect, preferred_direction="down"
) -> QRect:
    """
    Move/constrain the geometry for a drop down popup.

    Parameters
    ----------
    size : QSize
        The base popup size if not constrained.
    origin : QRect
        The origin rect from which the popup extends (in screen coords.).
    screen : QRect
        The available screen geometry into which the popup must fit.
    preferred_direction : str
        'up' or 'down'

    Returns
    -------
    geometry: QRect
        Constrained drop down list geometry to fit into screen
    """
    if preferred_direction == "down":
        # if the popup  geometry extends bellow the screen and there is more
        # room above the popup origin ...
        geometry = QRect(origin.bottomLeft() + QPoint(0, 1), size)
        if geometry.bottom() > screen.bottom() \
                and origin.center().y() > screen.center().y():
            # ...flip the rect about the origin so it extends upwards
            geometry.moveBottom(origin.top() - 1)
    elif preferred_direction == "up":
        geometry = QRect(origin.topLeft() - QPoint(0, 1 + size.height()), size)
        if geometry.top() < screen.top() \
                and origin.center().y() < screen.center().y():
            # ... flip, extend down
            geometry.moveTop(origin.bottom() - 1)
    else:
        raise ValueError(f"Invalid 'preferred_direction' ({preferred_direction})")

    # fixup horizontal position if it extends outside the screen
    if geometry.left() < screen.left():
        geometry.moveLeft(screen.left())
    if geometry.right() > screen.right():
        geometry.moveRight(screen.right())

    # bounded by screen geometry
    return geometry.intersected(screen)


def graphemes(text: str) -> Iterator[str]:
    """
    Return an iterator over grapheme clusters of text
    """
    # match internal QString encoding
    text_encoded = text.encode("utf-16-le")
    finder = QTextBoundaryFinder(QTextBoundaryFinder.Grapheme, text)
    start = 0
    while True:
        pos = finder.toNextBoundary()
        if pos == -1:
            return
        yield text_encoded[start*2: pos*2].decode("utf-16-le")
        start = pos


def grapheme_slice(text: str, start: int = 0, end: int = None) -> str:
    """
    Return a substring of text counting grapheme clusters not codepoints.
    """
    if start < 0 or (end is not None and end < 0):
        raise ValueError("negative start or end")

    s = 0
    slice_start: Optional[int] = None
    slice_end: Optional[int] = None
    for i, g in enumerate(graphemes(text)):
        if i == start:
            slice_start = s
        if i + 1 == end:
            slice_end = s + len(g)
            break
        s += len(g)
    if slice_start is None:
        return ""
    if slice_end is None:
        slice_end = len(text)
    return text[slice_start: slice_end]


def load_styled_icon(package: str, path: str) -> QIcon:
    return QIcon(StyledSvgIconEngine(pkgutil.get_data(package, path)))



"""Widget Settings and Settings Handlers

Settings are used to declare widget attributes that persist through sessions.
When widget is removed or saved to a schema file, its settings are packed,
serialized and stored. When a new widget is created, values of attributes
marked as settings are read from disk. When schema is loaded, attribute values
are set to one stored in schema.

Each widget has its own SettingsHandler that takes care of serializing and
storing of settings and SettingProvider that is incharge of reading and
writing the setting values.

All widgets extending from OWBaseWidget use SettingsHandler, unless they
declare otherwise. SettingsHandler ensures that setting attributes
are replaced with default (last used) setting values when the widget is
initialized and stored when the widget is removed.

Widgets with settings whose values depend on the widget inputs use
settings handlers based on ContextHandler. These handlers have two
additional methods, open_context and close_context.

open_context is called when widgets receives new data. It finds a suitable
context and sets the widget attributes to the values stored in context.
If no suitable context exists, a new one is created and values from widget
are copied to it.

close_context stores values that were last used on the widget to the context
so they can be used alter. It should be called before widget starts modifying
(initializing) the value of the setting attributes.
"""

import sys
import copy
import itertools
import os
import logging
import pickle
import pprint
import warnings
from operator import itemgetter
from typing import Any, Optional, Tuple

from orangewidget.gui import OWComponent

log = logging.getLogger(__name__)

__all__ = [
    "Setting", "SettingsHandler", "SettingProvider",
    "ContextSetting", "Context", "ContextHandler", "IncompatibleContext",
    "SettingsPrinter", "rename_setting", "widget_settings_dir"
]

_IMMUTABLES = (str, int, bytes, bool, float, tuple)

VERSION_KEY = "__version__"

# protocol v4 is supported since Python 3.4, protocol v5 since Python 3.8
PICKLE_PROTOCOL = 4


__WIDGET_SETTINGS_DIR = None  # type: Optional[Tuple[str, str]]


def set_widget_settings_dir_components(basedir: str, versionstr: str) -> None:
    """
    Set the directory path components where widgets save their settings.

    This overrides the global current application name/version derived paths.

    Note
    ----
    This should be set early in the application startup before any
    `OWBaseWidget` subclasses are imported (because it is needed at class
    definition time).

    Parameters
    ----------
    basedir: str
        The application specific data directory.
    versionstr: str
        The application version string for the versioned path component.

    See Also
    --------
    widget_settings_dir
    """
    global __WIDGET_SETTINGS_DIR
    __WIDGET_SETTINGS_DIR = (basedir, versionstr)


def widget_settings_dir(versioned=True) -> str:
    """
    Return the effective directory where widgets save their settings.

    This is a composed path based on a application specific data directory
    and application version string (if `versioned` is True) with a final
    'widgets' path component added (e.g. `'~/.local/share/MyApp/9.9.9/widgets'`)

    By default `QCoreApplication.applicationName` and
    `QCoreApplication.applicationVersion` are used to derive suitable paths
    (with a fallback if they are not set).

    Note
    ----
    If the application sets the `applicationName`/`applicationVersion`, it
    should do this early in the application startup before any
    `OWBaseWidget` subclasses are imported (because it is needed at class
    definition time).

    Use `set_widget_settings_dir_components` to override the default paths.

    Parameters
    ----------
    versioned: bool
        Should the returned path include the application version component.

    See Also
    --------
    set_widget_settings_dir_components
    """
    if __WIDGET_SETTINGS_DIR is None:
        from orangewidget.workflow.config import data_dir
        return os.path.join(data_dir(versioned), "widgets")
    else:
        base, version = __WIDGET_SETTINGS_DIR
        if versioned:
            return os.path.join(base, version, "widgets")
        else:
            return os.path.join(base, "widgets")


class Setting:
    """Description of a setting.
    """

    # Settings are automatically persisted to disk
    packable = True

    # Setting is only persisted to schema (default value does not change)
    schema_only = False

    def __new__(cls, default, *args, **kwargs):
        """A misleading docstring for providing type hints for Settings

        :type: default: T
        :rtype: T
        """
        return super().__new__(cls)

    def __init__(self, default, **data):
        self.name = None  # Name gets set in widget's meta class
        self.default = default
        self.__dict__.update(data)

    def __str__(self):
        return '{0} "{1}"'.format(self.__class__.__name__, self.name)

    __repr__ = __str__

    def __getnewargs__(self):
        return (self.default, )


# Pylint ignores type annotations in assignments. For
#
#    x: int = Setting(0)
#
# it ignores `int` and assumes x is of type `Setting`. Annotations in
# comments ( # type: int) also don't work. The only way to annotate x is
#
#    x: int
#    x = Setting(0)
#
# but we don't want to clutter the code with extra lines of annotations. Hence
# we disable checking the type of `Setting` by confusing pylint with an extra
# definition that is never executed.
if 1 == 0:
    class Setting:  # pylint: disable=function-redefined
        pass


def _apply_setting(setting: Setting, instance: OWComponent, value: Any):
    """
    Set `setting` of widget `instance` to the given `value`, in place if
    possible.

    If old and new values are of the same type, and the type is either a list
    or has methods `clear` and `update`, setting is updated in place. Otherwise
    the function calls `setattr`.

    Even if the target is update this way, call setattr (effectively calling
    instance.target = instance.target) to trigger any automatic updates related
    to this attribute).
    """
    target = getattr(instance, setting.name, None)
    if type(target) is type(value):
        if isinstance(value, list):
            target[:] = value
            value = target
        elif hasattr(value, "clear") and hasattr(value, "update"):
            target.clear()
            target.update(value)
            value = target
    setattr(instance, setting.name, value)


class SettingProvider:
    """A hierarchical structure keeping track of settings belonging to
    a class and child setting providers.

    At instantiation, it creates a dict of all Setting and SettingProvider
    members of the class. This dict is used to get/set values of settings
    from/to the instances of the class this provider belongs to.
    """

    def __init__(self, provider_class):
        """ Construct a new instance of SettingProvider.

        Traverse provider_class members and store all instances of
        Setting and SettingProvider.

        Parameters
        ----------
        provider_class : class
            class containing settings definitions
        """
        self.name = ""
        self.provider_class = provider_class
        self.providers = {}
        """:type: dict[str, SettingProvider]"""
        self.settings = {}
        """:type: dict[str, Setting]"""
        self.initialization_data = None

        for name in dir(provider_class):
            value = getattr(provider_class, name, None)
            if isinstance(value, Setting):
                value = copy.deepcopy(value)
                value.name = name
                self.settings[name] = value
            if isinstance(value, SettingProvider):
                value = copy.deepcopy(value)
                value.name = name
                self.providers[name] = value

    def initialize(self, instance, data=None):
        """Initialize instance settings to their default values.

        Mutable values are (shallow) copied before they are assigned to the
        widget. Immutable are used as-is.

        Parameters
        ----------
        instance : OWBaseWidget
            widget instance to initialize
        data : Optional[dict]
            optional data used to override the defaults
            (used when settings are loaded from schema)
        """
        if data is None and self.initialization_data is not None:
            data = self.initialization_data

        self._initialize_settings(instance, data)
        self._initialize_providers(instance, data)

    def reset_to_original(self, instance):
        self._initialize_settings(instance, None)
        self._initialize_providers(instance, None)

    def _initialize_settings(self, instance, data):
        if data is None:
            data = {}
        for name, setting in self.settings.items():
            value = data.get(name, setting.default)
            if isinstance(value, _IMMUTABLES):
                setattr(instance, name, value)
            else:
                setattr(instance, name, copy.copy(value))

    def _initialize_providers(self, instance, data):
        if not data:
            return

        for name, provider in self.providers.items():
            if name not in data:
                continue

            member = getattr(instance, name, None)
            if member is None or isinstance(member, SettingProvider):
                provider.store_initialization_data(data[name])
            else:
                provider.initialize(member, data[name])

    def store_initialization_data(self, initialization_data):
        """Store initialization data for later use.

        Used when settings handler is initialized, but member for this
        provider does not exists yet (because handler.initialize is called in
        __new__, but member will be created in __init__.

        Parameters
        ----------
        initialization_data : dict
            data to be used for initialization when the component is created
        """
        self.initialization_data = initialization_data

    @staticmethod
    def _default_packer(setting, instance):
        """A simple packet that yields setting name and value.

        Parameters
        ----------
        setting : Setting
        instance : OWBaseWidget
        """
        if setting.packable:
            if hasattr(instance, setting.name):
                yield setting.name, getattr(instance, setting.name)
            else:
                warnings.warn("{0} is declared as setting on {1} "
                              "but not present on instance."
                              .format(setting.name, instance))

    def pack(self, instance, packer=None):
        """Pack instance settings in a name:value dict.

        Parameters
        ----------
        instance : OWBaseWidget
            widget instance
        packer: callable (Setting, OWBaseWidget) -> Generator[(str, object)]
            optional packing function
            it will be called with setting and instance parameters and
            should yield (name, value) pairs that will be added to the
            packed_settings.
        """
        if packer is None:
            packer = self._default_packer

        packed_settings = dict(itertools.chain(
            *(packer(setting, instance) for setting in self.settings.values())
        ))

        packed_settings.update({
            name: provider.pack(getattr(instance, name), packer)
            for name, provider in self.providers.items()
            if hasattr(instance, name)
        })
        return packed_settings

    def unpack(self, instance, data):
        """Restore settings from data to the instance.

        Parameters
        ----------
        instance : OWBaseWidget
            instance to restore settings to
        data : dict
            packed data
        """
        for setting, _data, inst in self.traverse_settings(data, instance):
            if setting.name in _data and inst is not None:
                _apply_setting(setting, inst, _data[setting.name])

    def get_provider(self, provider_class):
        """Return provider for provider_class.

        If this provider matches, return it, otherwise pass
        the call to child providers.

        Parameters
        ----------
        provider_class : class
        """
        if issubclass(provider_class, self.provider_class):
            return self

        for subprovider in self.providers.values():
            provider = subprovider.get_provider(provider_class)
            if provider:
                return provider
        return None

    def traverse_settings(self, data=None, instance=None):
        """Generator of tuples (setting, data, instance) for each setting
        in this and child providers..

        Parameters
        ----------
        data : dict
            dictionary with setting values
        instance : OWBaseWidget
            instance matching setting_provider
        """
        data = data if data is not None else {}

        for setting in self.settings.values():
            yield setting, data, instance

        for provider in self.providers.values():
            data_ = data.get(provider.name, {})
            instance_ = getattr(instance, provider.name, None)
            for setting, component_data, component_instance in \
                    provider.traverse_settings(data_, instance_):
                yield setting, component_data, component_instance


class SettingsHandler:
    """Reads widget setting files and passes them to appropriate providers."""

    def __init__(self):
        """Create a setting handler template.

        Used in class definition. Bound instance will be created
        when SettingsHandler.create is called.
        """
        self.widget_class = None
        self.provider = None
        """:type: SettingProvider"""
        self.defaults = {}
        self.known_settings = {}

    @staticmethod
    def create(widget_class, template=None):
        """Create a new settings handler based on the template and bind it to
        widget_class.

        Parameters
        ----------
        widget_class : class
        template : SettingsHandler
            SettingsHandler to copy setup from

        Returns
        -------
        SettingsHandler
        """

        if template is None:
            template = SettingsHandler()

        setting_handler = copy.copy(template)
        setting_handler.defaults = {}
        setting_handler.bind(widget_class)
        return setting_handler

    def bind(self, widget_class):
        """Bind settings handler instance to widget_class.

        Parameters
        ----------
        widget_class : class
        """
        self.widget_class = widget_class
        self.provider = SettingProvider(widget_class)
        self.known_settings = {}
        self.analyze_settings(self.provider, "")
        self.read_defaults()

    def analyze_settings(self, provider, prefix):
        """Traverse through all settings known to the provider
        and analyze each of them.

        Parameters
        ----------
        provider : SettingProvider
        prefix : str
            prefix the provider is registered to handle
        """
        for setting in provider.settings.values():
            self.analyze_setting(prefix, setting)

        for name, sub_provider in provider.providers.items():
            new_prefix = '{0}{1}.'.format(prefix or '', name)
            self.analyze_settings(sub_provider, new_prefix)

    def analyze_setting(self, prefix, setting):
        """Perform any initialization task related to setting.

        Parameters
        ----------
        prefix : str
        setting : Setting
        """
        self.known_settings[prefix + setting.name] = setting

    def read_defaults(self):
        """Read (global) defaults for this widget class from a file.
        Opens a file and calls :obj:`read_defaults_file`. Derived classes
        should overload the latter."""
        filename = self._get_settings_filename()
        if os.path.isfile(filename):
            settings_file = open(filename, "rb")
            try:
                self.read_defaults_file(settings_file)
            # Unpickling exceptions can be of any type
            # pylint: disable=broad-except
            except Exception as ex:
                warnings.warn("Could not read defaults for widget {0}\n"
                              "The following error occurred:\n\n{1}"
                              .format(self.widget_class, ex))
            finally:
                settings_file.close()

    def read_defaults_file(self, settings_file):
        """Read (global) defaults for this widget class from a file.

        Parameters
        ----------
        settings_file : file-like object
        """
        defaults = pickle.load(settings_file)
        self.defaults = {
            key: value
            for key, value in defaults.items()
            if not isinstance(value, Setting)
        }
        self._migrate_settings(self.defaults)
        # remove schema_only settings introduced by the migration
        self._remove_schema_only(self.defaults)

    def write_defaults(self):
        """Write (global) defaults for this widget class to a file.
        Opens a file and calls :obj:`write_defaults_file`. Derived classes
        should overload the latter."""
        filename = self._get_settings_filename()
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        try:
            settings_file = open(filename, "wb")
            try:
                self.write_defaults_file(settings_file)
            except (EOFError, IOError, pickle.PicklingError) as ex:
                log.error("Could not write default settings for %s (%s).",
                          self.widget_class, ex)
                settings_file.close()
                os.remove(filename)
            else:
                settings_file.close()
        except PermissionError as ex:
            log.error("Could not write default settings for %s (%s).",
                      self.widget_class, type(ex).__name__)

    def write_defaults_file(self, settings_file):
        """Write defaults for this widget class to a file

        Parameters
        ----------
        settings_file : file-like object
        """
        defaults = dict(self.defaults)
        defaults[VERSION_KEY] = self.widget_class.settings_version
        pickle.dump(defaults, settings_file, protocol=PICKLE_PROTOCOL)

    def _get_settings_filename(self):
        """Return the name of the file with default settings for the widget"""
        return os.path.join(widget_settings_dir(),
                            "{0.__module__}.{0.__qualname__}.pickle"
                            .format(self.widget_class))

    def initialize(self, instance, data=None):
        """
        Initialize widget's settings.

        Replace all instance settings with their default values.

        Parameters
        ----------
        instance : OWBaseWidget
        data : dict or bytes that unpickle into a dict
            values used to override the defaults
        """
        provider = self._select_provider(instance)

        if isinstance(data, bytes):
            data = pickle.loads(data)
        self._migrate_settings(data)

        if provider is self.provider:
            data = self._add_defaults(data)

        provider.initialize(instance, data)

    def reset_to_original(self, instance):
        provider = self._select_provider(instance)
        provider.reset_to_original(instance)

    def _migrate_settings(self, settings):
        """Ask widget to migrate settings to the latest version."""
        if settings:
            try:
                self.widget_class.migrate_settings(
                    settings, settings.pop(VERSION_KEY, 0))
            except Exception:  # pylint: disable=broad-except
                sys.excepthook(*sys.exc_info())
                settings.clear()

    def _select_provider(self, instance):
        provider = self.provider.get_provider(instance.__class__)
        if provider is None:
            message = "{0} has not been declared as setting provider in {1}. " \
                      "Settings will not be saved/loaded properly. Defaults will be used instead." \
                      .format(instance.__class__, self.widget_class)
            warnings.warn(message)
            provider = SettingProvider(instance.__class__)
        return provider

    def _add_defaults(self, data):
        if data is None:
            return self.defaults

        new_data = self.defaults.copy()
        new_data.update(data)
        return new_data

    def _remove_schema_only(self, settings_dict):
        for setting, data, _ in self.provider.traverse_settings(data=settings_dict):
            if setting.schema_only:
                data.pop(setting.name, None)

    def _prepare_defaults(self, widget):
        self.defaults = self.provider.pack(widget)
        self._remove_schema_only(self.defaults)

    def pack_data(self, widget):
        """
        Pack the settings for the given widget. This method is used when
        saving schema, so that when the schema is reloaded the widget is
        initialized with its proper data and not the class-based defaults.
        See :obj:`SettingsHandler.initialize` for detailed explanation of its
        use.

        Inherited classes add other data, in particular widget-specific
        local contexts.

        Parameters
        ----------
        widget : OWBaseWidget
        """
        widget.settingsAboutToBePacked.emit()
        packed_settings = self.provider.pack(widget)
        packed_settings[VERSION_KEY] = self.widget_class.settings_version
        return packed_settings

    def update_defaults(self, widget):
        """
        Writes widget instance's settings to class defaults. Called when the
        widget is deleted.

        Parameters
        ----------
        widget : OWBaseWidget
        """
        widget.settingsAboutToBePacked.emit()
        self._prepare_defaults(widget)
        self.write_defaults()

    def fast_save(self, widget, name, value):
        """Store the (changed) widget's setting immediately to the context.

        Parameters
        ----------
        widget : OWBaseWidget
        name : str
        value : object

        """
        if name in self.known_settings:
            setting = self.known_settings[name]
            if not setting.schema_only:
                setting.default = value

    def reset_settings(self, instance):
        """Reset widget settings to defaults

        Parameters
        ----------
        instance : OWBaseWidget
        """
        for setting, _, inst in self.provider.traverse_settings(instance=instance):
            if setting.packable:
                _apply_setting(setting, inst, setting.default)


class ContextSetting(Setting):
    """Description of a context dependent setting"""

    OPTIONAL = 0
    REQUIRED = 2

    # Context settings are not persisted, but are stored in context instead.
    packable = False

    # These flags are not general - they assume that the setting has to do
    # something with the attributes. Large majority does, so this greatly
    # simplifies the declaration of settings in widget at no (visible)
    # cost to those settings that don't need it
    def __init__(self, default, *, required=2,
                 exclude_attributes=False, exclude_metas=False, **data):
        super().__init__(default, **data)
        self.exclude_attributes = exclude_attributes
        self.exclude_metas = exclude_metas
        self.required = required


class Context:
    """Class for data that defines context and
    values that should be applied to widget if given context
    is encountered."""
    def __init__(self, **argkw):
        self.values = {}
        self.__dict__.update(argkw)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class ContextHandler(SettingsHandler):
    """Base class for setting handlers that can handle contexts.

    Classes deriving from it need to implement method `match`.
    """

    NO_MATCH = 0
    MATCH = 1
    PERFECT_MATCH = 2

    MAX_SAVED_CONTEXTS = 50

    def __init__(self):
        super().__init__()
        self.global_contexts = []
        self.known_settings = {}

    def initialize(self, instance, data=None):
        """Initialize the widget: call the inherited initialization and
        add an attribute 'context_settings' to the widget. This method
        does not open a context."""
        instance.current_context = None
        super().initialize(instance, data)
        if data and "context_settings" in data:
            instance.context_settings = data["context_settings"]
            self._migrate_contexts(instance.context_settings)
        else:
            instance.context_settings = []

    def read_defaults_file(self, settings_file):
        """Call the inherited method, then read global context from the
           pickle."""
        super().read_defaults_file(settings_file)
        self.global_contexts = pickle.load(settings_file)
        self._migrate_contexts(self.global_contexts)
        # remove schema_only settings introduced by the migration
        for context in self.global_contexts:
            self._remove_schema_only(context.values)

    def _migrate_contexts(self, contexts):
        i = 0
        while i < len(contexts):
            context = contexts[i]
            try:
                self.widget_class.migrate_context(
                    context, context.values.pop(VERSION_KEY, 0))
            except IncompatibleContext:
                del contexts[i]
            except Exception:  # pylint: disable=broad-except
                sys.excepthook(*sys.exc_info())
                del contexts[i]
            else:
                i += 1

    def write_defaults_file(self, settings_file):
        """Call the inherited method, then add global context to the pickle."""
        super().write_defaults_file(settings_file)

        def add_version(context):
            context = copy.copy(context)
            context.values = dict(context.values)
            context.values[VERSION_KEY] = self.widget_class.settings_version
            return context

        pickle.dump([add_version(context) for context in self.global_contexts],
                    settings_file, protocol=PICKLE_PROTOCOL)

    def pack_data(self, widget):
        """Call the inherited method, then add local contexts to the dict."""
        data = super().pack_data(widget)
        self.settings_from_widget(widget)
        context_settings = [copy.copy(context) for context in
                            widget.context_settings]
        for context in context_settings:
            context.values[VERSION_KEY] = self.widget_class.settings_version
        data["context_settings"] = context_settings
        return data

    def update_defaults(self, widget):
        """
        Reimplemented from SettingsHandler

        Merge the widgets local contexts into the global contexts and persist
        the settings (including the contexts) to disk.
        """
        widget.settingsAboutToBePacked.emit()
        self.settings_from_widget(widget)
        globs = self.global_contexts
        assert widget.context_settings is not globs
        new_contexts = []
        for context in widget.context_settings:
            context = copy.deepcopy(context)
            self._remove_schema_only(context.values)
            if context not in globs:
                new_contexts.append(context)
        globs[:0] = reversed(new_contexts)
        del globs[self.MAX_SAVED_CONTEXTS:]

        # Save non-context settings. Do not call super().update_defaults, so that
        # settingsAboutToBePacked is emitted once.
        self._prepare_defaults(widget)
        self.write_defaults()

    def new_context(self, *args):
        """Create a new context."""
        return Context()

    def open_context(self, widget, *args):
        """Open a context by finding one and setting the widget data or
        creating one and fill with the data from the widget."""
        widget.current_context, is_new = \
            self.find_or_create_context(widget, *args)
        if is_new:
            self.settings_from_widget(widget, *args)
        else:
            self.settings_to_widget(widget, *args)

    def match(self, context, *args):
        """Return the degree to which the stored `context` matches the data
        passed in additional arguments).
        When match returns 0 (ContextHandler.NO_MATCH), the context will not
        be used. When it returns ContextHandler.PERFECT_MATCH, the context
        is a perfect match so no further search is necessary.

        If imperfect matching is not desired, match should only
        return ContextHandler.NO_MATCH or ContextHandler.PERFECT_MATCH.

        Derived classes must overload this method.
        """
        raise NotImplementedError

    def find_or_create_context(self, widget, *args):
        """Find the best matching context or create a new one if nothing
        useful is found. The returned context is moved to or added to the top
        of the context list."""

        # First search the contexts that were already used in this widget instance
        best_context, best_score = self.find_context(widget.context_settings, args, move_up=True)
        # If the exact data was used, reuse the context
        if best_score == self.PERFECT_MATCH:
            return best_context, False

        # Otherwise check if a better match is available in global_contexts
        best_context, best_score = self.find_context(self.global_contexts, args,
                                                     best_score, best_context)
        if best_context:
            context = self.clone_context(best_context, *args)
        else:
            context = self.new_context(*args)
        # Store context in widget instance. It will be pushed to global_contexts
        # when (if) update defaults is called.
        self.add_context(widget.context_settings, context)
        return context, best_context is None

    def find_context(self, known_contexts, args, best_score=0, best_context=None, move_up=False):
        """Search the given list of contexts and return the context
         which best matches the given args.

        best_score and best_context can be used to provide base_values.
        """

        best_idx = None
        for i, context in enumerate(known_contexts):
            score = self.match(context, *args)
            if score > best_score:  # NO_MATCH is not OK!
                best_context, best_score, best_idx = context, score, i
                if score == self.PERFECT_MATCH:
                    break
        if best_idx is not None and move_up:
            self.move_context_up(known_contexts, best_idx)
        return best_context, best_score

    @staticmethod
    def move_context_up(contexts, index):
        """Move the context to the top of the list"""
        contexts.insert(0, contexts.pop(index))

    def add_context(self, contexts, setting):
        """Add the context to the top of the list."""
        contexts.insert(0, setting)
        del contexts[self.MAX_SAVED_CONTEXTS:]

    def clone_context(self, old_context, *args):
        """Construct a copy of the context settings suitable for the context
        described by additional arguments. The method is called by
        find_or_create_context with the same arguments. A class that overloads
        :obj:`match` to accept additional arguments must also overload
        :obj:`clone_context`."""
        context = self.new_context(*args)
        context.values = copy.deepcopy(old_context.values)

        traverse = self.provider.traverse_settings
        for setting, data, _ in traverse(data=context.values):
            if not isinstance(setting, ContextSetting):
                continue

            self.filter_value(setting, data, *args)
        return context

    @staticmethod
    def filter_value(setting, data, *args):
        """Remove values related to setting that are invalid given args."""

    def close_context(self, widget):
        """Close the context by calling :obj:`settings_from_widget` to write
        any relevant widget settings to the context."""
        if widget.current_context is None:
            return

        self.settings_from_widget(widget)
        widget.current_context = None

    def settings_to_widget(self, widget, *args):
        """Apply context settings stored in currently opened context
        to the widget.
        """
        context = widget.current_context
        if context is None:
            return

        widget.retrieveSpecificSettings()

        for setting, data, instance in \
                self.provider.traverse_settings(data=context.values, instance=widget):
            if not isinstance(setting, ContextSetting) or setting.name not in data:
                continue
            value = self.decode_setting(setting, data[setting.name], *args)
            _apply_setting(setting, instance, value)

    def settings_from_widget(self, widget, *args):
        """Update the current context with the setting values from the widget.
        """

        context = widget.current_context
        if context is None:
            return

        widget.storeSpecificSettings()

        def packer(setting, instance):
            if isinstance(setting, ContextSetting) and hasattr(instance, setting.name):
                value = getattr(instance, setting.name)
                yield setting.name, self.encode_setting(context, setting, value)

        context.values = self.provider.pack(widget, packer=packer)

    def fast_save(self, widget, name, value):
        """Update value of `name` setting in the current context to `value`
        """
        setting = self.known_settings.get(name)
        if isinstance(setting, ContextSetting):
            context = widget.current_context
            if setting.schema_only or context is None:
                return

            value = self.encode_setting(context, setting, value)
            self.update_packed_data(context.values, name, value)
        else:
            super().fast_save(widget, name, value)

    @staticmethod
    def update_packed_data(data, name, value):
        """Updates setting value stored in data dict"""

        *prefixes, name = name.split('.')
        for prefix in prefixes:
            data = data.setdefault(prefix, {})
        data[name] = value

    def encode_setting(self, context, setting, value):
        """Encode value to be stored in settings dict"""
        return copy.copy(value)

    def decode_setting(self, setting, value, *args):
        """Decode settings value from the setting dict format"""
        return value


class IncompatibleContext(Exception):
    """Raised when a required variable in context is not available in data."""


class SettingsPrinter(pprint.PrettyPrinter):
    """Pretty Printer that knows how to properly format Contexts."""

    def _format(self, obj, stream, indent, allowance, context, level):
        if not isinstance(obj, Context):
            super()._format(obj, stream, indent, allowance, context, level)
            return

        stream.write("Context(")
        for key, value in sorted(obj.__dict__.items(), key=itemgetter(0)):
            if key == "values":
                continue
            stream.write(key)
            stream.write("=")
            stream.write(self._repr(value, context, level + 1))
            stream.write(",\n")
            stream.write(" " * (indent + 8))
        stream.write("values=")
        stream.write(" ")
        self._format(obj.values, stream, indent+15,
                     allowance+1, context, level + 1)
        stream.write(")")


def rename_setting(settings, old_name, new_name):
    """
    Rename setting from `old_name` to `new_name`. Used in migrations.

    The argument `settings` can be `dict` or `Context`.
    """
    if isinstance(settings, Context):
        rename_setting(settings.values, old_name, new_name)
    else:
        settings[new_name] = settings.pop(old_name)



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring
from orangewidget import gui
from orangewidget.gui import OWComponent
from orangewidget.settings import SettingProvider
from orangewidget.tests.base import WidgetTest
from orangewidget.widget import OWBaseWidget


class DummyComponent(OWComponent):
    foo = True


class MyWidget(OWBaseWidget):
    foo = True

    component = SettingProvider(DummyComponent)

    def __init__(self):
        super().__init__()

        self.component = DummyComponent(self)
        self.foo_control = gui.checkBox(self.controlArea, self, "foo", "")
        self.component_foo_control = \
            gui.checkBox(self.controlArea, self, "component.foo", "")


class ControlGetterTests(WidgetTest):
    def test_getter(self):
        widget = self.create_widget(MyWidget)
        self.assertIs(widget.controls.foo, widget.foo_control)
        self.assertIs(widget.controls.component.foo,
                      widget.component_foo_control)



from itertools import chain

import numpy as np

from matplotlib.colors import to_hex
from pyqtgraph.graphicsItems.ScatterPlotItem import ScatterPlotItem
from AnyQt.QtCore import Qt
from AnyQt.QtGui import QPen, QBrush


def numpy_repr(a):
    """ A numpy repr without summarization """
    opts = np.get_printoptions()
    # avoid numpy repr as it changes between versions
    # TODO handle numpy repr differences
    if isinstance(a, np.ndarray):
        return "array(" + repr(a.tolist()) + ")"
    try:
        np.set_printoptions(threshold=10**10)
        return repr(a)
    finally:
        np.set_printoptions(**opts)


def numpy_repr_int(a):
    # avoid numpy repr as it changes between versions
    # TODO handle numpy repr differences
    if isinstance(a, np.ndarray):
        a = a.tolist()
    else:
        a = list(a)
    return "array(" + repr(a) + ", dtype='int')"


def compress_if_all_same(l):
    s = set(l)
    if len(s) == 1:
        v = s.pop()
        return v.item() if isinstance(v, np.generic) else v
    else:
        return l


def is_sequence_not_string(a):
    if isinstance(a, str):
        return False
    try:
        iter(a)
        return True
    except TypeError:
        pass
    return False


def code_with_indices(data, data_name, indices, indices_name):
    if is_sequence_not_string(data) and indices is not None:
        return data_name + "[" + indices_name + "]"
    else:
        return data_name


def index_per_different(l):
    different = []
    different_ind = {}
    index = []
    for e in l:
        if e not in different_ind:
            different_ind[e] = len(different)
            different.append(e)
        index.append(different_ind[e])
    return different, index


def scatterplot_code(scatterplot_item):
    x = scatterplot_item.data['x']
    y = scatterplot_item.data['y']
    sizes = scatterplot_item.data["size"]

    code = []

    code.append("# data")
    code.append("x = {}".format(numpy_repr(x)))
    code.append("y = {}".format(numpy_repr(y)))

    code.append("# style")
    sizes = compress_if_all_same(sizes)
    if sizes == -1:
        sizes = None
    code.append("sizes = {}".format(numpy_repr(sizes)))

    def colortuple(pen):
        if isinstance(pen, (QPen, QBrush)):
            color = pen.color()
            return color.redF(), color.greenF(), color.blueF(), color.alphaF()
        return pen

    def width(pen):
        if isinstance(pen, QPen):
            return pen.widthF()
        return pen

    linewidths = np.array([width(a) for a in scatterplot_item.data["pen"]])

    def shown(a):
        if isinstance(a, (QPen, QBrush)):
            s = a.style()
            if s == Qt.NoPen or s == Qt.NoBrush or a.color().alpha() == 0:
                return False
        return True

    shown_edge = [shown(a) for a in scatterplot_item.data["pen"]]
    shown_brush = [shown(a) for a in scatterplot_item.data["brush"]]

    # return early if the scatterplot is all transparent
    if not any(shown_edge) and not any(shown_brush):
        return ""

    def do_colors(code, data_column, show, name):
        colors = [colortuple(a) for a in data_column]
        if all(a is None for a in colors):
            colors, index = None, None
        else:
            # replace None values with blue colors
            colors = np.array([((0, 0, 1, 1) if a is None else a)
                               for a in colors])
            # set alpha for hidden (Qt.NoPen, Qt.NoBrush) elements to zero
            colors[:, 3][np.array(show) == 0] = 0
            # shorter color names for printout
            colors = [to_hex(c, keep_alpha=True) for c in colors]
            colors, index = index_per_different(colors)

        code.append("{} = {}".format(name, repr(colors)))
        if index is not None:
            code.append("{}_index = {}".format(name, numpy_repr_int(index)))

        decompresssed_code = name
        if index is not None:
            decompresssed_code = "array({})[{}_index]".format(name, name)
            colors = np.array(colors)[index]

        return colors, decompresssed_code

    edgecolors, edgecolors_code = do_colors(code, scatterplot_item.data["pen"], shown_edge, "edgecolors")
    facecolors, facecolors_code = do_colors(code, scatterplot_item.data["brush"], shown_brush, "facecolors")

    linewidths = compress_if_all_same(linewidths)
    code.append("linewidths = {}".format(numpy_repr(linewidths)))

    # possible_markers for scatterplot are in .graph.CurveSymbols
    def matplotlib_marker(m):
        if m == "t":
            return "^"
        elif m == "t2":
            return ">"
        elif m == "t3":
            return "<"
        elif m == "star":
            return "*"
        elif m == "+":
            return "P"
        elif m == "x":
            return "X"
        return m

    # TODO labels are missing

    # each marker requires one call to matplotlib's scatter!
    markers = np.array([matplotlib_marker(m) for m in scatterplot_item.data["symbol"]])
    for m in set(markers):
        indices = np.where(markers == m)[0]
        if np.all(indices == np.arange(x.shape[0])):
            indices = None
        if indices is not None:
            code.append("indices = {}".format(numpy_repr_int(indices)))

        def indexed(data, data_name, indices=indices):
            return code_with_indices(data, data_name, indices, "indices")

        code.append("plt.scatter(x={}, y={}, s={}, marker={},\n"
                    "            facecolors={}, edgecolors={},\n"
                    "            linewidths={})"
                    .format(indexed(x, "x"),
                            indexed(y, "y"),
                            (indexed(sizes, "sizes") + "**2/4") if sizes is not None else "sizes",
                            repr(m),
                            indexed(facecolors, facecolors_code),
                            indexed(edgecolors, edgecolors_code),
                            indexed(linewidths, "linewidths")
                           ))

    return "\n".join(code)


def scene_code(scene):

    code = []

    code.append("import matplotlib.pyplot as plt")
    code.append("import numpy as np")
    code.append("from numpy import array")

    code.append("")
    code.append("plt.clf()")

    code.append("")

    for item in scene.items:
        if isinstance(item, ScatterPlotItem):
            code.append(scatterplot_code(item))

    # TODO currently does not work for graphs without axes and for multiple axes!
    for position, set_ticks, set_label in [("bottom", "plt.xticks", "plt.xlabel"),
                                           ("left", "plt.yticks", "plt.ylabel")]:
        axis = scene.getAxis(position)
        code.append("{}({})".format(set_label, repr(str(axis.labelText))))

        # textual tick labels
        if axis._tickLevels is not None:
            major_minor = list(chain(*axis._tickLevels))
            locs = [a[0] for a in major_minor]
            labels = [a[1] for a in major_minor]
            code.append("{}({}, {})".format(set_ticks, locs, repr(labels)))

    return "\n".join(code)



from orangecanvas.registry import WidgetDescription
from orangecanvas.registry import discovery


def widget_desc_from_module(module):
    """
    Get the widget description from a module.

    The module is inspected for classes that have a method
    `get_widget_description`. The function calls this method and expects
    a dictionary, which is used as keyword arguments for
    :obj:`WidgetDescription`. This method also converts all signal types
    into qualified names to prevent import problems when cached
    descriptions are unpickled (the relevant code using this lists should
    be able to handle missing types better).

    Parameters
    ----------
    module (`module` or `str`): a module to inspect

    Returns
    -------
    An instance of :obj:`WidgetDescription`
    """
    if isinstance(module, str):
        module = __import__(module, fromlist=[""])

    for widget_class in module.__dict__.values():
        if not hasattr(widget_class, "get_widget_description"):
            continue
        description = widget_class.get_widget_description()
        if description is None:
            continue

        description = WidgetDescription(**description)
        description.package = module.__package__
        description.category = widget_class.category
        return description

    raise discovery.WidgetSpecificationError


class WidgetDiscovery(discovery.WidgetDiscovery):

    def widget_description(self, module, widget_name=None, category_name=None,
                           distribution=None):
        """
        Return widget description from a module.
        """
        module = discovery.asmodule(module)
        desc = widget_desc_from_module(module)

        if widget_name is not None:
            desc.name = widget_name

        if category_name is not None and desc.category is None:
            desc.category = category_name

        if distribution is not None:
            desc.project_name = distribution.name

        return desc



import os
import io
import logging
import traceback
import warnings
import pickle
import pkgutil
from collections import OrderedDict
from enum import IntEnum

from typing import Optional

from AnyQt.QtCore import Qt, QObject, pyqtSlot, QSize
from AnyQt.QtGui import (
    QIcon, QCursor, QStandardItemModel, QStandardItem, QPageSize)
from AnyQt.QtWidgets import (
    QApplication, QDialog, QFileDialog, QTableView, QHeaderView,
    QMessageBox)
from AnyQt.QtPrintSupport import QPrinter, QPrintDialog


from orangewidget import gui
from orangewidget.utils import load_styled_icon
from orangewidget.widget import OWBaseWidget
from orangewidget.settings import Setting

# Importing WebviewWidget can fail if neither QWebKit or QWebEngine
# are available
try:
    from orangewidget.utils.webview import WebviewWidget
except ImportError:  # pragma: no cover
    WebviewWidget = None
    HAVE_REPORT = False
else:
    HAVE_REPORT = True


log = logging.getLogger(__name__)

class Column(IntEnum):
    item = 0
    remove = 1
    scheme = 2


class ReportItem(QStandardItem):
    def __init__(self, name, html, scheme, module, icon_name, comment=""):
        self.name = name
        self.html = html
        self.scheme = scheme
        self.module = module
        self.icon_name = icon_name
        self.comment = comment
        try:
            icon = load_styled_icon(module, icon_name)
        except (ImportError, FileNotFoundError):
            icon = QIcon()
        self.id = id(icon)
        super().__init__(icon, name)

    def __getnewargs__(self):
        return (self.name, self.html, self.scheme, self.module, self.icon_name,
                self.comment)


class ReportItemModel(QStandardItemModel):
    def __init__(self, rows, columns, parent=None):
        super().__init__(rows, columns, parent)

    def add_item(self, item):
        row = self.rowCount()
        self.setItem(row, Column.item, item)
        self.setItem(row, Column.remove, self._icon_item("Remove"))
        self.setItem(row, Column.scheme, self._icon_item("Open Scheme"))

    def get_item_by_id(self, item_id):
        for i in range(self.rowCount()):
            item = self.item(i)
            if str(item.id) == item_id:
                return item
        return None

    @staticmethod
    def _icon_item(tooltip):
        item = QStandardItem()
        item.setEditable(False)
        item.setToolTip(tooltip)
        return item


class ReportTable(QTableView):
    def __init__(self, parent):
        super().__init__(parent)
        self._icon_remove = load_styled_icon(__package__, "icons/delete.svg")
        self._icon_scheme = load_styled_icon(__package__, "icons/scheme.svg")

    def mouseMoveEvent(self, event):
        self._clear_icons()
        self._repaint(self.indexAt(event.pos()))

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            super().mouseReleaseEvent(event)
        self._clear_icons()
        self._repaint(self.indexAt(event.pos()))

    def leaveEvent(self, _):
        self._clear_icons()

    def _repaint(self, index):
        row, column = index.row(), index.column()
        if column in (Column.remove, Column.scheme):
            self.setCursor(QCursor(Qt.PointingHandCursor))
        else:
            self.setCursor(QCursor(Qt.ArrowCursor))
        if row >= 0:
            self.model().item(row, Column.remove).setIcon(self._icon_remove)
            self.model().item(row, Column.scheme).setIcon(self._icon_scheme)

    def _clear_icons(self):
        model = self.model()
        for i in range(model.rowCount()):
            model.item(i, Column.remove).setIcon(QIcon())
            model.item(i, Column.scheme).setIcon(QIcon())


class OWReport(OWBaseWidget):
    name = "Report"
    save_dir = Setting("")
    open_dir = Setting("")

    def __init__(self):
        super().__init__()
        self._setup_ui_()
        self.report_changed = False
        self.have_report_warning_shown = False

        index_file = pkgutil.get_data(__name__, "index.html")
        self.report_html_template = index_file.decode("utf-8")

    def _setup_ui_(self):
        self.table_model = ReportItemModel(0, len(Column.__members__))
        self.table = ReportTable(self.controlArea)
        self.table.setModel(self.table_model)
        self.table.setShowGrid(False)
        self.table.setSelectionBehavior(QTableView.SelectRows)
        self.table.setSelectionMode(QTableView.SingleSelection)
        self.table.setWordWrap(False)
        self.table.setMouseTracking(True)
        self.table.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.table.verticalHeader().setDefaultSectionSize(20)
        self.table.verticalHeader().setVisible(False)
        self.table.horizontalHeader().setVisible(False)
        self.table.setFixedWidth(250)
        self.table.setColumnWidth(Column.item, 200)
        self.table.setColumnWidth(Column.remove, 23)
        self.table.setColumnWidth(Column.scheme, 25)
        self.table.clicked.connect(self._table_clicked)
        self.table.selectionModel().selectionChanged.connect(
            self._table_selection_changed)
        self.controlArea.layout().addWidget(self.table)

        self.last_scheme = None
        self.scheme_button = gui.button(
            self.controlArea, self, "Back to Last Scheme",
            callback=self._show_last_scheme
        )
        box = gui.hBox(self.controlArea)
        box.setContentsMargins(-6, 0, -6, 0)
        self.save_button = gui.button(
            box, self, "Save", callback=self.save_report, disabled=True
        )
        self.print_button = gui.button(
            box, self, "Print", callback=self._print_report, disabled=True
        )

        class PyBridge(QObject):
            @pyqtSlot(str)
            def _select_item(myself, item_id):
                item = self.table_model.get_item_by_id(item_id)
                self.table.selectRow(self.table_model.indexFromItem(item).row())
                self._change_selected_item(item)

            @pyqtSlot(str, str)
            def _add_comment(myself, item_id, value):
                item = self.table_model.get_item_by_id(item_id)
                item.comment = value
                self.report_changed = True

        if WebviewWidget is not None:
            self.report_view = WebviewWidget(self.mainArea, bridge=PyBridge(self))
            self.mainArea.layout().addWidget(self.report_view)
        else:
            self.report_view = None

    def sizeHint(self):
        return QSize(850, 500)

    def _table_clicked(self, index):
        if index.column() == Column.remove:
            self._remove_item(index.row())
            indexes = self.table.selectionModel().selectedIndexes()
            if indexes:
                item = self.table_model.item(indexes[0].row())
                self._scroll_to_item(item)
                self._change_selected_item(item)
        if index.column() == Column.scheme:
            self._show_scheme(index.row())

    def _table_selection_changed(self, new_selection, _):
        if new_selection.indexes():
            item = self.table_model.item(new_selection.indexes()[0].row())
            self._scroll_to_item(item)
            self._change_selected_item(item)

    def _remove_item(self, row):
        self.table_model.removeRow(row)
        self._empty_report()
        self.report_changed = True
        self._build_html()

    def clear(self):
        self.table_model.clear()
        self._empty_report()
        self.report_changed = True
        self._build_html()

    def _add_item(self, widget):
        name = widget.get_widget_name_extension()
        name = "{} - {}".format(widget.name, name) if name else widget.name
        item = ReportItem(name, widget.report_html, self._get_scheme(),
                          widget.__module__, widget.icon)
        self.table_model.add_item(item)
        self._empty_report()
        self.report_changed = True
        return item

    def _empty_report(self):
        # disable save and print if no reports
        self.save_button.setEnabled(self.table_model.rowCount())
        self.print_button.setEnabled(self.table_model.rowCount())

    def _build_html(self, selected_id=None):
        if not self.report_view:
            return
        html = self.report_html_template
        if selected_id is not None:
            onload = f"(function (id) {{" \
                     f"     setSelectedId(id); scrollToId(id); " \
                     f"}}" \
                     f"(\"{selected_id}\"));" \
                     f""
            html += f"<body onload='{onload}'>"
        else:
            html += "<body>"
        for i in range(self.table_model.rowCount()):
            item = self.table_model.item(i)
            html += "<div id='{}' class='normal' " \
                    "onClick='pybridge._select_item(this.id)'>{}<div " \
                    "class='textwrapper'><textarea " \
                    "placeholder='Write a comment...'" \
                    "onInput='this.innerHTML = this.value;" \
                    "pybridge._add_comment(this.parentNode.parentNode.id, this.value);'" \
                    ">{}</textarea></div>" \
                    "</div>".format(item.id, item.html, item.comment)
        html += "</body></html>"
        self.report_view.setHtml(html)

    def _scroll_to_item(self, item):
        if not self.report_view:
            return
        self.report_view.runJavaScript(
            f"scrollToId('{item.id}')",
            lambda res: log.debug("scrollToId returned %s", res)
        )

    def _change_selected_item(self, item):
        if not self.report_view:
            return
        self.report_view.runJavaScript(
            f"setSelectedId('{item.id}');",
            lambda res: log.debug("setSelectedId returned %s", res)
        )
        self.report_changed = True

    def make_report(self, widget):
        item = self._add_item(widget)
        self._build_html(item.id)
        self.table.selectionModel().selectionChanged.disconnect(
            self._table_selection_changed
        )
        self.table.selectRow(self.table_model.rowCount() - 1)
        self.table.selectionModel().selectionChanged.connect(
            self._table_selection_changed
        )

    def _get_scheme(self):
        canvas = self.get_canvas_instance()
        if canvas is None:
            return None
        scheme = canvas.current_document().scheme()
        return self._get_scheme_str(scheme)

    def _get_scheme_str(self, scheme):
        buffer = io.BytesIO()
        scheme.save_to(buffer, pickle_fallback=True)
        return buffer.getvalue().decode("utf-8")

    def _show_scheme(self, row):
        scheme = self.table_model.item(row).scheme
        canvas = self.get_canvas_instance()
        if canvas is None:
            return
        document = canvas.current_document()
        if document.isModifiedStrict():
            self.last_scheme = self._get_scheme_str(document.scheme())
        self._load_scheme(scheme)

    def _show_last_scheme(self):
        if self.last_scheme:
            self._load_scheme(self.last_scheme)

    def _load_scheme(self, contents):
        # forcibly load the contents into the associated CanvasMainWindow
        # instance if one exists. Preserve `self` as the designated report.
        canvas = self.get_canvas_instance()
        if canvas is not None:
            document = canvas.current_document()
            scheme = document.scheme()
            # Clear the undo stack as it will no longer apply to the new
            # workflow.
            document.undoStack().clear()
            scheme.clear()
            scheme.load_from(io.StringIO(contents))

    def save_report(self):
        """Save report"""
        formats = (('HTML (*.html)', '.html'),
                   ('PDF (*.pdf)', '.pdf')) if self.report_view else tuple()
        formats = formats + (('Report (*.report)', '.report'),)
        formats = OrderedDict(formats)

        filename, selected_format = QFileDialog.getSaveFileName(
            self, "Save Report", self.save_dir, ';;'.join(formats.keys()))
        if not filename:
            return QDialog.Rejected

        # Set appropriate extension if not set by the user
        expect_ext = formats[selected_format]
        if not filename.endswith(expect_ext):
            filename += expect_ext

        self.save_dir = os.path.dirname(filename)
        self.saveSettings()
        _, extension = os.path.splitext(filename)
        if extension == ".pdf":
            printer = QPrinter()
            printer.setPageSize(QPageSize(QPageSize.A4))
            printer.setOutputFormat(QPrinter.PdfFormat)
            printer.setOutputFileName(filename)
            self._print_to_printer(printer)
        elif extension == ".report":
            self.save(filename)
        else:
            def save_html(contents):
                try:
                    with open(filename, "w", encoding="utf-8") as f:
                        f.write(contents)
                except PermissionError:
                    self.permission_error(filename)

            if self.report_view:
                save_html(self.report_view.html())
        self.report_changed = False
        return QDialog.Accepted

    def _print_to_printer(self, printer):
        if not self.report_view:
            return
        filename = printer.outputFileName()
        if filename:
            try:
                # QtWebEngine
                return self.report_view.page().printToPdf(filename)
            except AttributeError:
                try:
                    # QtWebKit
                    return self.report_view.print_(printer)
                except AttributeError:
                    # QtWebEngine 5.6
                    pass
        # Fallback to printing widget as an image
        self.report_view.render(printer)

    def _print_report(self):
        printer = QPrinter()
        print_dialog = QPrintDialog(printer, self)
        print_dialog.setWindowTitle("Print report")
        if print_dialog.exec() != QDialog.Accepted:
            return
        self._print_to_printer(printer)

    def save(self, filename):
        attributes = {}
        for key in ('last_scheme', 'open_dir'):
            attributes[key] = getattr(self, key, None)
        items = [self.table_model.item(i)
                 for i in range(self.table_model.rowCount())]
        report = dict(__version__=1,
                      attributes=attributes,
                      items=items)

        try:
            with open(filename, 'wb') as f:
                pickle.dump(report, f)
        except PermissionError:
            self.permission_error(filename)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            report = pickle.load(f)

        if not isinstance(report, dict):
            return report

        self = cls()
        self.__dict__.update(report['attributes'])
        for item in report['items']:
            self.table_model.add_item(
                ReportItem(item.name, item.html, item.scheme,
                           item.module, item.icon_name, item.comment)
            )
        return self

    def permission_error(self, filename):
        log.error("PermissionError when trying to write report.", exc_info=True)
        mb = QMessageBox(
            self,
            icon=QMessageBox.Critical,
            windowTitle=self.tr("Error"),
            text=self.tr("Permission error when trying to write report."),
            informativeText=self.tr("Permission error occurred "
                                    "while saving '{}'.").format(filename),
            detailedText=traceback.format_exc(limit=20)
        )
        mb.setWindowModality(Qt.WindowModal)
        mb.setAttribute(Qt.WA_DeleteOnClose)
        mb.exec()

    def is_empty(self):
        return not self.table_model.rowCount()

    def is_changed(self):
        return self.report_changed

    @staticmethod
    def set_instance(report):
        warnings.warn(
            "OWReport.set_instance is deprecated",
            DeprecationWarning, stacklevel=2
        )
        app_inst = QApplication.instance()
        app_inst._report_window = report

    @staticmethod
    def get_instance():
        warnings.warn(
            "OWReport.get_instance is deprecated",
            DeprecationWarning, stacklevel=2
        )
        app_inst = QApplication.instance()
        if not hasattr(app_inst, "_report_window"):
            report = OWReport()
            app_inst._report_window = report
        return app_inst._report_window

    def get_canvas_instance(self):
        # type: () -> Optional[CanvasMainWindow]
        """
        Return a CanvasMainWindow instance to which this report is attached.

        Return None if not associated with any window.

        Returns
        -------
        window : Optional[CanvasMainWindow]
        """
        try:
            from orangewidget.workflow.mainwindow import OWCanvasMainWindow
        except ImportError:
            return None
        # Run up the parent/window chain
        parent = self.parent()
        if parent is not None:
            window = parent.window()
            if isinstance(window, OWCanvasMainWindow):
                return window

    def copy_to_clipboard(self):
        if self.report_view:
            self.report_view.triggerPageAction(self.report_view.page().Copy)




# THIS FILE IS GENERATED FROM ORANGE-WIDGET-BASE SETUP.PY
short_version = '4.25.1'
version = '4.25.1'
full_version = '4.25.1'
git_revision = '0392606b2a76682fa53193b40da7d6c6ef42d987'
release = True

if not release:
    version = full_version
    short_version += ".dev"



import unittest
from unittest.mock import Mock, patch

from orangewidget.tests.base import GuiTest


class SkipTest(Exception):
    pass


class TestGuiTest(unittest.TestCase):
    @patch("unittest.case.SkipTest", SkipTest)
    def test_english(self):
        class TestA(GuiTest):
            pure_test = Mock()
            test = GuiTest.skipNonEnglish(pure_test)

            pure_test_si = Mock()
            test_si = GuiTest.runOnLanguage("Slovenian")(pure_test_si)

        test_obj = TestA()

        test_obj.test()
        test_obj.pure_test.assert_called()

        self.assertRaises(SkipTest, test_obj.test_si)
        test_obj.pure_test_si.assert_not_called()

    @patch("unittest.case.SkipTest", SkipTest)
    @patch("orangewidget.tests.base.GuiTest.LANGUAGE", "Slovenian")
    def test_non_english(self):
        class TestA(GuiTest):
            pure_test = Mock()
            test = GuiTest.skipNonEnglish(pure_test)

            pure_test_si = Mock()
            test_si = GuiTest.runOnLanguage("Slovenian")(pure_test_si)

            pure_test_fr = Mock()
            test_fr = GuiTest.runOnLanguage("French")(pure_test_fr)

        test_obj = TestA()

        self.assertRaises(SkipTest, test_obj.test)
        test_obj.pure_test.assert_not_called()

        test_obj.test_si()
        test_obj.pure_test_si.assert_called()

        self.assertRaises(SkipTest, test_obj.test_fr)
        test_obj.pure_test_fr.assert_not_called()


if __name__ == "__main__":
    unittest.main()



import os
import sys
import tempfile
from collections import OrderedDict
from typing import Union

import numpy as np

from AnyQt import QtGui, QtCore, QtSvg, QtWidgets
from AnyQt.QtCore import QMarginsF, Qt, QRectF, QPoint, QRect, QSize
from AnyQt.QtGui import QPalette
from AnyQt.QtWidgets import (
    QGraphicsScene, QGraphicsView, QApplication, QWidget
)

from orangewidget.utils.matplotlib_export import scene_code

try:
    from orangewidget.utils.webview import WebviewWidget
except ImportError:
    WebviewWidget = None

# This is needed just for type annotation
try:
    from pyqtgraph import GraphicsItem
except:
    class GraphicsItem:
        pass

__all__ = [
    "ImgFormat", "Compression", "PngFormat", "ClipboardFormat", "SvgFormat",
    "MatplotlibPDFFormat", "MatplotlibFormat", "PdfFormat",
]


class Compression:
    """Supported compression extensions"""
    GZIP = '.gz'
    BZIP2 = '.bz2'
    XZ = '.xz'
    all = (GZIP, BZIP2, XZ)


class _Registry(type):
    """Metaclass that registers subtypes."""
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        if not hasattr(cls, 'registry'):
            cls.registry = OrderedDict()
        else:
            cls.registry[name] = cls
        return cls

    def __iter__(cls):
        return iter(cls.registry)

    def __str__(cls):
        if cls in cls.registry.values():
            return cls.__name__
        return '{}({{{}}})'.format(cls.__name__, ', '.join(cls.registry))


def effective_background(scene: QGraphicsScene, view: QGraphicsView):
    background = scene.backgroundBrush()
    if background.style() != Qt.NoBrush:
        return background
    background = view.backgroundBrush()
    if background.style() != Qt.NoBrush:
        return background
    viewport = view.viewport()
    role = viewport.backgroundRole()
    if role != QPalette.NoRole:
        return viewport.palette().brush(role)
    return viewport.palette().brush(QPalette.Window)


class classproperty(property):
    def __get__(self, instance, class_):
        return self.fget(class_)


class ImgFormat(metaclass=_Registry):
    PRIORITY = sys.maxsize

    @staticmethod
    def _get_buffer(size, filename):
        raise NotImplementedError

    @staticmethod
    def _get_target(source):
        return QtCore.QRectF(0, 0, source.width(), source.height())

    @classmethod
    def _setup_painter(cls, painter, object, source_rect, buffer):
        pass

    @staticmethod
    def _save_buffer(buffer, filename):
        raise NotImplementedError

    @staticmethod
    def _meta_data(buffer):
        meta_data = {}

        try:
            size = buffer.size()
        except AttributeError:
            pass
        else:
            meta_data["width"] = size.width()
            meta_data["height"] = size.height()

        try:
            meta_data["pixel_ratio"] = buffer.devicePixelRatio()
        except AttributeError:
            pass

        return meta_data

    @staticmethod
    def _get_exporter():
        raise NotImplementedError

    @staticmethod
    def _export(self, exporter, filename):
        raise NotImplementedError

    @classmethod
    def write_image(
            cls, filename,
            object: Union[GraphicsItem,  # via save_pyqtgraph
                          QGraphicsScene,  # via save_scene
                          QGraphicsView,  # via save_widget
                          QWidget  # via save_widget, but with different render
                          ]):
        def get_scene_pixel_ratio(scene: QGraphicsScene):
            views = scene.views()
            if views:
                return views[0].devicePixelRatio()
            try:
                # It is unusual for scene not to be viewed, except in tests.
                # As a fallback, we get ratio for (any) screen
                return QApplication.primaryScreen().devicePixelRatio()
            except:  # pylint: disable=broad-except
                # If there is no screen (in tests on headless server?) assume 1;
                # the worst that can happen is low resolution of images
                return 1

        def save_pyqtgraph():
            assert isinstance(object, GraphicsItem)
            exporter = cls._get_exporter()
            scene = object.scene()
            if scene is None:
                return cls._export(exporter(scene), filename)
            views = scene.views()
            if views:
                # preserve scene rect and background brush
                scenerect = scene.sceneRect()
                backgroundbrush = scene.backgroundBrush()
                try:
                    view = scene.views()[0]
                    scene.setSceneRect(view.sceneRect())
                    scene.setBackgroundBrush(effective_background(scene, view))
                    return cls._export(exporter(scene), filename)
                finally:
                    # reset scene rect and background brush
                    scene.setBackgroundBrush(backgroundbrush)
                    scene.setSceneRect(scenerect)
            else:
                return cls._export(exporter(scene), filename)

        def save_scene():
            assert isinstance(object, QGraphicsScene)
            ratio = get_scene_pixel_ratio(object)
            views = object.views()
            if not views:
                rect = object.itemsBoundingRect()
                return _render(rect, ratio, rect.size(), object)

            # Pick the first view. If there's a widget with multiple views that
            # cares which one is used, it must set graph_name to view, not scene
            view = views[0]
            rect = view.sceneRect()
            target_rect = view.mapFromScene(rect).boundingRect()
            source_rect = QRect(
                int(target_rect.x()), int(target_rect.y()),
                int(target_rect.width()), int(target_rect.height()))
            return _render(source_rect, ratio, target_rect.size(), view)

        def save_widget():
            assert isinstance(object, QWidget)
            return _render(object.rect(), object.devicePixelRatio(), object.size(),
                    object)

        def _render(
                source_rect: QRectF, pixel_ratio: float, size: QSize,
                renderer: Union[QGraphicsScene, QGraphicsView, QWidget]):
            buffer_size = size + type(size)(30, 30)
            try:
                buffer = cls._get_buffer(buffer_size, filename, pixel_ratio)
            except TypeError:  # backward compatibility (with what?)
                buffer = cls._get_buffer(buffer_size, filename)

            painter = QtGui.QPainter()
            painter.begin(buffer)
            try:
                painter.setRenderHint(QtGui.QPainter.Antialiasing)
                if QtCore.QT_VERSION >= 0x050D00:
                    painter.setRenderHint(QtGui.QPainter.LosslessImageRendering)
                cls._setup_painter(
                    painter, renderer,
                    QRectF(0, 0, buffer_size.width(), buffer_size.height()), buffer)
                if isinstance(renderer, (QGraphicsView, QGraphicsScene)):
                    renderer.render(painter,
                                    QRectF(15, 15, size.width(), size.height()),
                                    source_rect)
                else:
                    assert isinstance(object, QWidget)
                    renderer.render(painter, QPoint(15, 15))
            finally:
                # In case of exception, end painting so that we get an exception
                # not a core dump
                painter.end()
            cls._save_buffer(buffer, filename)
            return cls._meta_data(buffer)

        if isinstance(object, GraphicsItem):
            return save_pyqtgraph()
        elif isinstance(object, QGraphicsScene):
            return save_scene()
        elif isinstance(object, QWidget):  # this includes QGraphicsView
            return save_widget()
        else:
            raise TypeError(f"{cls.__name__} "
                            f"cannot imagine {type(object).__name__}")

    @classmethod
    def write(cls, filename, scene):
        if type(scene) == dict:
            scene = scene['scene']
        return cls.write_image(filename, scene)

    @classproperty
    def img_writers(cls):  # type: () -> Mapping[str, Type[ImgFormat]]
        formats = OrderedDict()
        for format in sorted(cls.registry.values(), key=lambda x: x.PRIORITY):
            for ext in getattr(format, 'EXTENSIONS', []):
                # Only adds if not yet registered
                formats.setdefault(ext, format)
        return formats

    graph_writers = img_writers

    @classproperty
    def formats(cls):
        return cls.registry.values()


class PngFormat(ImgFormat):
    EXTENSIONS = ('.png',)
    DESCRIPTION = 'Portable Network Graphics'
    PRIORITY = 50

    @staticmethod
    def _get_buffer(size, filename, ratio=1):
        img = QtGui.QPixmap(int(size.width() * ratio),
                            int(size.height() * ratio))
        img.setDevicePixelRatio(ratio)
        return img

    @classmethod
    def _setup_painter(cls, painter, object, source_rect, buffer):
        if isinstance(object, (QGraphicsScene, QGraphicsView)):
            brush = object.backgroundBrush()
            if brush.style() == QtCore.Qt.NoBrush:
                brush = QtGui.QBrush(object.palette().color(QtGui.QPalette.Base))
        else:
            brush = QtGui.QBrush(QtCore.Qt.white)
        painter.fillRect(source_rect, brush)

    @staticmethod
    def _save_buffer(buffer, filename):
        image = buffer.toImage()
        dpm = int(2835 * image.devicePixelRatio())
        image.setDotsPerMeterX(dpm)
        image.setDotsPerMeterY(dpm)
        image.save(filename, "png")

    @staticmethod
    def _get_exporter():
        from pyqtgraph.exporters.ImageExporter import ImageExporter
        from pyqtgraph import functions as fn

        # Use devicePixelRatio
        class PngExporter(ImageExporter):
            def __init__(self, item):
                super().__init__(item)
                if isinstance(item, QGraphicsScene):
                    self.ratio = item.views()[0].devicePixelRatio()
                else:
                    # Let's hope it's a view or another QWidget
                    self.ratio = item.devicePixelRatio()

            # Copied verbatim from super;
            # changes are in three lines that define self.png
            # and lines related to setDotsPerMeterX after painter.end()
            def export(self, fileName=None, toBytes=False, copy=False):
                if fileName is None and not toBytes and not copy:
                    filter = self.getSupportedImageFormats()
                    self.fileSaveDialog(filter=filter)
                    return

                w = int(self.params['width'])
                h = int(self.params['height'])
                if w == 0 or h == 0:
                    raise Exception(
                        "Cannot export image with size=0 (requested "
                        "export size is %dx%d)" % (w, h))

                targetRect = QtCore.QRect(0, 0, w, h)
                sourceRect = self.getSourceRect()

                self.png = QtGui.QImage(
                    int(w * self.ratio), int(h * self.ratio),
                    QtGui.QImage.Format.Format_ARGB32)
                self.png.fill(self.params['background'])
                self.png.setDevicePixelRatio(self.ratio)

                ## set resolution of image:
                origTargetRect = self.getTargetRect()
                resolutionScale = targetRect.width() / origTargetRect.width()
                # self.png.setDotsPerMeterX(self.png.dotsPerMeterX() * resolutionScale)
                # self.png.setDotsPerMeterY(self.png.dotsPerMeterY() * resolutionScale)

                painter = QtGui.QPainter(self.png)
                # dtr = painter.deviceTransform()
                try:
                    self.setExportMode(True, {
                        'antialias': self.params['antialias'],
                        'background': self.params['background'],
                        'painter': painter,
                        'resolutionScale': resolutionScale})
                    painter.setRenderHint(
                        QtGui.QPainter.RenderHint.Antialiasing,
                        self.params['antialias'])
                    self.getScene().render(painter, QtCore.QRectF(targetRect),
                                           QtCore.QRectF(sourceRect))
                finally:
                    self.setExportMode(False)
                painter.end()

                # setDotsPerMeter* after painting avoids missized axes problem
                dpm = int(2835 * self.ratio)
                self.png.setDotsPerMeterX(dpm)
                self.png.setDotsPerMeterY(dpm)

                if self.params['invertValue']:
                    bg = fn.ndarray_from_qimage(self.png)
                    if sys.byteorder == 'little':
                        cv = slice(0, 3)
                    else:
                        cv = slice(1, 4)
                    mn = bg[..., cv].min(axis=2)
                    mx = bg[..., cv].max(axis=2)
                    d = (255 - mx) - mn
                    bg[..., cv] += d[..., np.newaxis]

                if copy:
                    QtWidgets.QApplication.clipboard().setImage(self.png)
                elif toBytes:
                    return self.png
                else:
                    return self.png.save(fileName)

        return PngExporter

    @classmethod
    def _export(cls, exporter, filename):
        buffer = exporter.export(toBytes=True)
        buffer.save(filename, "png")
        return cls._meta_data(buffer)


class ClipboardFormat(PngFormat):
    EXTENSIONS = ()
    DESCRIPTION = 'System Clipboard'
    PRIORITY = 50

    @classmethod
    def _save_buffer(cls, buffer, _):
        meta_data = cls._meta_data(buffer)
        image = buffer.toImage()
        if meta_data is not None:
            ratio = meta_data.get("pixel_ratio", 1)
            dpm = int(2835 * ratio)
            image.setDotsPerMeterX(dpm)
            image.setDotsPerMeterY(dpm)
        QApplication.clipboard().setImage(image)

    @staticmethod
    def _export(exporter, _):
        exporter.export(copy=True)


class SvgFormat(ImgFormat):
    EXTENSIONS = ('.svg',)
    DESCRIPTION = 'Scalable Vector Graphics'
    PRIORITY = 100

    @staticmethod
    def _get_buffer(size, filename):
        buffer = QtSvg.QSvgGenerator()
        buffer.setResolution(int(QApplication.primaryScreen().logicalDotsPerInch()))
        buffer.setFileName(filename)
        buffer.setViewBox(QtCore.QRectF(0, 0, size.width(), size.height()))
        return buffer

    @staticmethod
    def _save_buffer(buffer, filename):
        dev = buffer.outputDevice()
        if dev is not None:
            dev.flush()
        pass

    @staticmethod
    def _get_exporter():
        from pyqtgraph.exporters.SVGExporter import SVGExporter
        return SVGExporter

    @staticmethod
    def _export(exporter, filename):
        if isinstance(exporter.item, QGraphicsScene):
            scene = exporter.item
            params = exporter.parameters()
            brush = effective_background(scene, scene.views()[0])
            params.param("background").setValue(brush.color())
        exporter.export(filename)

    @classmethod
    def write_image(cls, filename, scene):
        # WebviewWidget exposes its SVG contents more directly;
        # no need to go via QPainter if we can avoid it
        svg = None
        if WebviewWidget is not None and isinstance(scene, WebviewWidget):
            try:
                svg = scene.svg()
            except (ValueError, IOError):
                pass
        if svg is None:
            super().write_image(filename, scene)
            svg = open(filename).read()
        svg = svg.replace(
            "<svg ",
            '<svg style="image-rendering:optimizeSpeed;image-rendering:pixelated" ')
        with open(filename, 'w') as f:
            f.write(svg)


class MatplotlibFormat:
    # not registered as a FileFormat as it only works with scatter plot
    EXTENSIONS = ('.py',)
    DESCRIPTION = 'Python Code (with Matplotlib)'
    PRIORITY = 300

    @classmethod
    def write_image(cls, filename, scene):
        code = scene_code(scene) + "\n\nplt.show()"
        with open(filename, "wt") as f:
            f.write(code)

    @classmethod
    def write(cls, filename, scene):
        if type(scene) == dict:
            scene = scene['scene']
        cls.write_image(filename, scene)


class MatplotlibPDFFormat(MatplotlibFormat):
    EXTENSIONS = ('.pdf',)
    DESCRIPTION = 'Portable Document Format (from Matplotlib)'
    PRIORITY = 200

    @classmethod
    def write_image(cls, filename, scene):
        import matplotlib
        with matplotlib.rc_context({"backend": "pdf"}):
            code = scene_code(scene) + "\n\nplt.savefig({})".format(repr(filename))
            exec(code, {})  # will generate a pdf


if QtCore.QT_VERSION >= 0x050C00:  # Qt 5.12+

    class PdfFormat(ImgFormat):
        EXTENSIONS = ('.pdf', )
        DESCRIPTION = 'Portable Document Format'
        PRIORITY = 110

        @staticmethod
        def _get_buffer(size, filename):
            buffer = QtGui.QPdfWriter(filename)
            dpi = int(QApplication.primaryScreen().logicalDotsPerInch())
            buffer.setResolution(dpi)
            buffer.setPageMargins(QMarginsF(0, 0, 0, 0))
            pagesize = QtCore.QSizeF(size.width(), size.height()) / dpi * 25.4
            buffer.setPageSize(QtGui.QPageSize(pagesize, QtGui.QPageSize.Millimeter))
            return buffer

        @staticmethod
        def _save_buffer(buffer, filename):
            pass

        @staticmethod
        def _get_exporter():
            from orangewidget.utils.PDFExporter import PDFExporter
            return PDFExporter

        @staticmethod
        def _export(exporter, filename):
            exporter.export(filename)

else:

    # older Qt version have PdfWriter bugs and are handled through SVG

    class PdfFormat(ImgFormat):
        EXTENSIONS = ('.pdf', )
        DESCRIPTION = 'Portable Document Format'
        PRIORITY = 110

        @classmethod
        def write_image(cls, filename, scene):
            # export via svg to temp file then print that
            # NOTE: can't use NamedTemporaryFile with delete = True
            # (see https://bugs.python.org/issue14243)
            fd, tmpname = tempfile.mkstemp(suffix=".svg")
            os.close(fd)
            try:
                SvgFormat.write_image(tmpname, scene)
                with open(tmpname, "rb") as f:
                    svgcontents = f.read()
            finally:
                os.unlink(tmpname)

            svgrend = QtSvg.QSvgRenderer(QtCore.QByteArray(svgcontents))
            vbox = svgrend.viewBox()
            if not vbox.isValid():
                size = svgrend.defaultSize()
            else:
                size = vbox.size()
            writer = QtGui.QPdfWriter(filename)
            pagesize = QtGui.QPageSize(QtCore.QSizeF(size) * 0.282,
                                       QtGui.QPageSize.Millimeter)
            writer.setPageSize(pagesize)
            painter = QtGui.QPainter(writer)
            svgrend.render(painter)
            painter.end()
            del svgrend
            del painter



import os
from typing import Optional

from AnyQt.QtCore import Qt, QSettings, QTimer
from AnyQt.QtWidgets import (
    QAction, QFileDialog, QMenu, QMenuBar, QWidget, QMessageBox, QDialog,
    QApplication)
from AnyQt.QtGui import QKeySequence

from orangecanvas.application.canvasmain import CanvasMainWindow
from orangewidget.report.owreport import HAVE_REPORT, OWReport
from orangewidget.workflow.widgetsscheme import WidgetsScheme


def _insert_action(mb, menuid, beforeactionid, action):
    # type: (QMenuBar, str, str, QAction) -> bool
    """
    Insert an action into one of a QMenuBar's menu.

    Parameters
    ----------
    mb : QMenuBar
        The menu bar
    menuid : str
        The target menu's objectName. The menu must be a child of `mb`.
    beforeactionid : str
        The objectName of the action before which the action will be inserted.
    action : QAction
        The action to insert

    Returns
    -------
    success: bool
        True if the actions was successfully inserted (the menu and before
        actions were found), False otherwise
    """
    def find_action(widget, name):  # type: (QWidget, str) -> Optional[QAction]
        for a in widget.actions():
            if a.objectName() == name:
                return a
        return None

    menu = mb.findChild(QMenu, menuid)
    if menu is not None:
        sep = find_action(menu, beforeactionid)
        if sep:
            menu.insertAction(sep, action)
            return True
    return False


class OWCanvasMainWindow(CanvasMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show_report_action = QAction(
            "Show report", self,
            objectName="action-show-report",
            toolTip="Show a report window",
            shortcut=QKeySequence("Shift+R"),
            enabled=HAVE_REPORT,
        )
        self.show_report_action.triggered.connect(self.show_report_view)
        self.open_report_action = QAction(
            "Open Report...", self,
            objectName="action-open-report",
            toolTip="Open a saved report",
            enabled=HAVE_REPORT,
        )
        self.open_report_action.triggered.connect(self.open_report)
        self.reset_widget_settings_action = QAction(
            self.tr("Reset Widget Settings..."), self,
            triggered=self.reset_widget_settings
        )

        menubar = self.menuBar()
        # Insert the 'Load report' in the File menu ...
        _insert_action(menubar, "file-menu", "open-actions-separator",
                       self.open_report_action)
        # ... and 'Show report' in the View menu.
        _insert_action(menubar, "view-menu", "view-visible-actions-separator",
                       self.show_report_action)

        _insert_action(menubar, "options-menu", "canvas-addons-action",
                       self.reset_widget_settings_action)

    def open_report(self):
        """
        Present an 'Open report' dialog to the user, load a '.report' file
        (as saved by OWReport) and create a new canvas window associated
        with the OWReport instance.
        """
        settings = QSettings()
        KEY = "report/file-dialog-dir"
        start_dir = settings.value(KEY, "", type=str)
        dlg = QFileDialog(
            self,
            windowTitle=self.tr("Open Report"),
            acceptMode=QFileDialog.AcceptOpen,
            fileMode=QFileDialog.ExistingFile,
        )
        if os.path.isdir(start_dir):
            dlg.setDirectory(start_dir)

        dlg.setWindowModality(Qt.ApplicationModal)
        dlg.setNameFilters(["Report (*.report)"])

        def accepted():
            directory = dlg.directory().absolutePath()
            filename = dlg.selectedFiles()[0]
            settings.setValue(KEY, directory)
            self._open_report(filename)

        dlg.accepted.connect(accepted)
        dlg.exec()

    def _open_report(self, filename):
        """
        Open and load a '*.report' from 'filename'
        """
        report = OWReport.load(filename)
        # Create a new window for the report
        if self.is_transient():
            window = self
        else:
            window = self.create_new_window()
        # toggle the window modified flag (this will clear the 'is_transient'
        # flag on the new window)
        window.setWindowModified(True)
        window.setWindowModified(False)

        report.setParent(window, Qt.Window)
        sc = window.current_document().scheme()  # type: WidgetsScheme
        sc.set_report_view(report)

        window.show()
        window.raise_()
        window.show_report_view()

        report._build_html()
        report.table.selectRow(0)
        report.show()
        report.raise_()

    def show_report_view(self):
        """
        Show the 'Report' view for the current workflow.
        """
        sc = self.current_document().scheme()  # type: WidgetsScheme
        sc.show_report_view()

    def reset_widget_settings(self):
        name = QApplication.applicationName() or 'Orange'
        mb = QMessageBox(
            self,
            windowTitle="Clear settings",
            text=f"{name} needs to be restarted for the changes to take effect.",
            icon=QMessageBox.Information,
            informativeText=f"Press OK to restart {name} now.",
            standardButtons=QMessageBox.Ok | QMessageBox.Cancel,
        )
        res = mb.exec()
        if res == QMessageBox.Ok:
            # Touch a finely crafted file inside the settings directory.
            # The existence of this file is checked by the canvas main
            # function and is deleted there.
            from orangewidget.settings import widget_settings_dir
            dirname = widget_settings_dir()
            try:
                os.makedirs(dirname, exist_ok=True)
            except (FileExistsError, PermissionError):
                return
            with open(os.path.join(dirname, "DELETE_ON_START"), "a"):
                pass

            def restart():
                quit_temp_val = QApplication.quitOnLastWindowClosed()
                QApplication.setQuitOnLastWindowClosed(False)
                QApplication.closeAllWindows()
                windows = QApplication.topLevelWindows()
                if any(w.isVisible() for w in windows):  # if a window close was cancelled
                    QApplication.setQuitOnLastWindowClosed(quit_temp_val)
                    QMessageBox(
                        text="Restart Cancelled",
                        informativeText=f"Settings will be reset on {name}'s next restart",
                        icon=QMessageBox.Information
                    ).exec()
                else:
                    QApplication.exit(96)

            QTimer.singleShot(0, restart)

    def ask_save_report(self):
        """
        Ask whether to save the report or not.
        Returns:
            `QDialog.Rejected` if user cancels, `QDialog.Accepted` otherwise
        """
        workflow = self.current_document().scheme()
        if not isinstance(workflow, WidgetsScheme) or not workflow.has_report():
            return QDialog.Accepted

        report = workflow.report_view()
        if not report.is_changed():
            return QDialog.Accepted

        mBox = QMessageBox(
            self,
            windowTitle="Report window",
            icon=QMessageBox.Question,
            text="The report contains unsaved changes.",
            informativeText="Would you like to save the report?",
            standardButtons=QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel,
        )
        mBox.setDefaultButton(QMessageBox.Save)
        answ = mBox.exec()
        if answ == QMessageBox.Cancel:
            return QDialog.Rejected
        if answ == QMessageBox.Save:
            return report.save_report()
        return QDialog.Accepted

    def closeEvent(self, event):
        if self.ask_save_report() == QDialog.Rejected:
            event.ignore()
            return
        super().closeEvent(event)



import sys

import os
import unittest
from unittest.mock import Mock
from tempfile import NamedTemporaryFile

from orangewidget.utils.filedialogs import RecentPath, open_filename_dialog, \
    unambiguous_paths


class TestUtils(unittest.TestCase):
    def test_unambiguous_paths(self):
        paths = [
            "asd.txt",
            "abc/def/ghi.txt",
            "abc/def/jkl.txt",
            "abc/xyz/jkl.txt",
            "abc/xyz/rty/qwe.txt",
            "abd/xyz/rty/qwe.txt",
            "abe/xyz/rty/qwe.txt",
        ]

        paths = [t.replace("/", os.path.sep, 1) for t in paths]

        def test(exp, **kwargs):
            self.assertEqual(unambiguous_paths(paths, **kwargs),
                             [t.replace("/", os.path.sep) for t in exp])

        test(["asd.txt",
              "ghi.txt",
              "def/jkl.txt",
              "xyz/jkl.txt",
              "abc/xyz/rty/qwe.txt",
              "abd/xyz/rty/qwe.txt",
              "abe/xyz/rty/qwe.txt"])

        test(["asd.txt",
              "def/ghi.txt",
              "def/jkl.txt",
              "xyz/jkl.txt",
              "abc/xyz/rty/qwe.txt",
              "abd/xyz/rty/qwe.txt",
              "abe/xyz/rty/qwe.txt"], minlevel=2)

        test(["asd.txt",
              "abc/def/ghi.txt",
              "abc/def/jkl.txt",
              "abc/xyz/jkl.txt",
              "abc/xyz/rty/qwe.txt",
              "abd/xyz/rty/qwe.txt",
              "abe/xyz/rty/qwe.txt"], minlevel=3)

        test(["asd.txt",
              "abc/def/ghi.txt",
              "abc/def/jkl.txt",
              "abc/xyz/jkl.txt",
              "abc/xyz/rty/qwe.txt",
              "abd/xyz/rty/qwe.txt",
              "abe/xyz/rty/qwe.txt"], minlevel=4)

        # For simplicity, omit this test on Windows; if it works on Posix paths,
        # it works on Windows, too.
        if os.path.sep == "/":
            t = ["abc/def/ghi.txt", "abc/def/ghi.txt"]
            self.assertEqual(unambiguous_paths(t), t)

    if sys.platform == "win32":
        ptch1 = ptch2 = ptch3 = lambda x: x
    else:
        # This test is intended for Windows, but for easier testing of a test
        # on non-Windows machine, we patch it to make it work on others
        ptch1 = unittest.mock.patch("os.path.sep", "/")
        ptch2 = unittest.mock.patch("os.path.altsep", "\\")
        ptch3 = unittest.mock.patch(
            "os.path.join",
            lambda *args, oj=os.path.join: oj(*args).replace("/", "\\"))
    @ptch1
    @ptch2
    @ptch3
    def test_unambiguous_paths_windows(self):
        paths = ["C:\\Documents/Newsletters\\Summer2018.pdf",
                 "D:\\Documents/Newsletters\\Summer2018.pdf"]
        self.assertEqual(unambiguous_paths(paths),
                         ["C:\\Documents\\Newsletters\\Summer2018.pdf",
                          "D:\\Documents\\Newsletters\\Summer2018.pdf"]
                         )

        paths = ["C:\\abc\\def\\Summer2018.pdf",
                 "C:\\abc\\deg\\Summer2018.pdf"]
        self.assertEqual(unambiguous_paths(paths),
                         ["def\\Summer2018.pdf",
                          "deg\\Summer2018.pdf", ]
                         )

        paths = ["C:\\deg\\Summer2018.pdf",
                 "D:\\deg\\Summer2018.pdf"]
        self.assertEqual(unambiguous_paths(paths),
                         ["C:\\deg\\Summer2018.pdf",
                          "D:\\deg\\Summer2018.pdf", ]
                         )


class TestRecentPath(unittest.TestCase):
    def test_resolve(self):
        temp_file = NamedTemporaryFile(dir=os.getcwd(), delete=False)
        file_name = temp_file.name
        temp_file.close()
        base_name = os.path.basename(file_name)
        try:
            recent_path = RecentPath(
                os.path.join("temp/datasets", base_name), "",
                os.path.join("datasets", base_name)
            )
            search_paths = [("basedir", os.getcwd())]
            self.assertIsNotNone(recent_path.resolve(search_paths))
        finally:
            os.remove(file_name)


class TestOpenFilenameDialog(unittest.TestCase):
    def test_empty_filter(self):
        class ABCFormat:
            EXTENSIONS = ('.abc', '.jkl')
            DESCRIPTION = 'abc file'
            PRIORITY = 30

        name, file_format, file_filter = open_filename_dialog(
            ".", "", [ABCFormat],
            dialog=Mock(return_value=("foo.xyz", "")))
        self.assertEqual(name, "foo.xyz")
        self.assertEqual(file_format, None)
        self.assertEqual(file_filter, None)

    def test_add_all(self):
        class ABCFormat:
            EXTENSIONS = ('.abc', '.jkl')
            DESCRIPTION = 'abc file'
            PRIORITY = 30

        class DEFFormat:
            EXTENSIONS = ('.def', )
            DESCRIPTION = 'def file'
            PRIORITY = 40

        # Add all known extensions
        dialog = Mock(return_value=("foo.abc", ""))
        open_filename_dialog(".", "", [ABCFormat, DEFFormat], dialog=dialog)
        self.assertIn("(*.abc *.def *.jkl)", dialog.call_args[0][3])
        self.assertIn("(*.abc *.jkl)", dialog.call_args[0][3])
        self.assertIn("(*.def)", dialog.call_args[0][3])

        # Add all extensions (*.*)
        dialog = Mock(return_value=("foo.abc", ""))
        open_filename_dialog(".", "", [ABCFormat, DEFFormat],
                             add_all="*", dialog=dialog)
        self.assertIn("(*.*)", dialog.call_args[0][3])
        self.assertIn("(*.abc *.jkl)", dialog.call_args[0][3])
        self.assertIn("(*.def)", dialog.call_args[0][3])

        # Don't add any extensions
        dialog = Mock(return_value=("foo.abc", ""))
        open_filename_dialog(".", "", [ABCFormat, DEFFormat],
                             add_all=False, dialog=dialog)
        self.assertNotIn("(*.abc *.def *.jkl)", dialog.call_args[0][3])
        self.assertIn("(*.abc *.jkl)", dialog.call_args[0][3])
        self.assertIn("(*.def)", dialog.call_args[0][3])

        # With a single format, add all known extensions is ignored
        dialog = Mock(return_value=("foo.abc", ""))
        open_filename_dialog(".", "", [ABCFormat],
                             dialog=dialog)
        self.assertNotIn("(*.*)", dialog.call_args[0][3])
        self.assertEqual(dialog.call_args[0][3].count("(*.abc *.jkl)"), 1)

        # With a single format, add all extensions (*.*) still applies
        dialog = Mock(return_value=("foo.abc", ""))
        open_filename_dialog(".", "", [ABCFormat],
                             add_all="*", dialog=dialog)
        self.assertIn("(*.*)", dialog.call_args[0][3])
        self.assertIn("(*.abc *.jkl)", dialog.call_args[0][3])


if __name__ == "__main__":
    unittest.main()



"""
Base (example) Configuration for an Orange Widget based application.
"""
import warnings

import os
import sys
import itertools
from typing import Dict, Any, Optional, Iterable, List

import requests

from AnyQt.QtCore import QStandardPaths, QCoreApplication

from orangecanvas import config
from orangecanvas.utils.pkgmeta import entry_points, EntryPoint


# generated from biolab/orange3-addons repository
OFFICIAL_ADDON_LIST = "https://orange.biolab.si/addons/list"

WIDGETS_ENTRY = "orange.widgets"


class Config(config.Default):
    """
    Basic configuration for running orangewidget based workflow application.
    """
    OrganizationDomain = "biolab.si"
    ApplicationName = "Orange Canvas"
    try:
        from orangewidget.version import short_version as ApplicationVersion
    except ImportError:
        ApplicationVersion = "0.0.0"

    @staticmethod
    def widgets_entry_points():
        """
        Return an `EntryPoint` iterator for all registered 'orange.widgets'
        entry points.
        """
        return entry_points(group=WIDGETS_ENTRY)

    @staticmethod
    def addon_entry_points():
        return Config.widgets_entry_points()

    @staticmethod
    def addon_defaults_list(session=None):
        # type: (Optional[requests.Session]) -> List[Dict[str, Any]]
        """
        Return a list of available add-ons.
        """
        if session is None:
            session = requests.Session()
        return session.get(OFFICIAL_ADDON_LIST).json()

    @staticmethod
    def core_packages():
        # type: () -> List[str]
        """
        Return a list of 'core packages'

        These packages constitute required the application framework. They
        cannot be removes via the 'Add-on/plugins' manager. They however can
        be updated. The package that defines the application's `main()` entry
        point must always be in this list.
        """
        return ["orange-widget-base"]

    @staticmethod
    def examples_entry_points():
        # type: () -> Iterable[EntryPoint]
        """
        Return an iterator over the entry points yielding 'Example Workflows'
        """
        return entry_points(group="orange.widgets.tutorials")

    @staticmethod
    def widget_discovery(*args, **kwargs):
        from .discovery import WidgetDiscovery
        return WidgetDiscovery(*args, **kwargs)

    @staticmethod
    def workflow_constructor(*args, **kwargs):
        from .widgetsscheme import WidgetsScheme
        return WidgetsScheme(*args, **kwargs)


def data_dir_base():
    """
    Return the platform dependent generic application directory.

    This is usually

        - on windows: "%USERPROFILE%\\AppData\\Local\\"
        - on OSX:  "~/Library/Application Support/"
        - other: "~/.local/share/
    """
    return QStandardPaths.writableLocation(QStandardPaths.GenericDataLocation)


def data_dir(versioned=True):
    """
    Return the platform dependent application data directory.

    This is ``data_dir_base()``/{NAME}/{VERSION}/ directory if versioned is
    `True` and ``data_dir_base()``/{NAME}/ otherwise, where NAME is
    `QCoreApplication.applicationName()` and VERSION is
    `QCoreApplication.applicationVersion()`.
    """
    base = data_dir_base()
    assert base
    name = QCoreApplication.applicationName()
    version = QCoreApplication.applicationVersion()
    if not name:
        name = "Orange"
    if not version:
        version = "0.0.0"
    if versioned:
        return os.path.join(base, name, version)
    else:
        return os.path.join(base, name)


def cache_dir():
    """Return the application cache directory. If the directory path
    does not yet exists then create it.
    """
    warnings.warn(
        f"'{__name__}.cache_dir' is deprecated.",
        DeprecationWarning, stacklevel=2
    )
    base = QStandardPaths.writableLocation(QStandardPaths.GenericCacheLocation)
    name = QCoreApplication.applicationName()
    version = QCoreApplication.applicationVersion()
    if not name:
        name = "Orange"
    if not version:
        version = "0.0.0"
    path = os.path.join(base, name, version)
    try:
        os.makedirs(path, exist_ok=True)
    except OSError:
        pass
    return path


def log_dir():
    """
    Return the application log directory.
    """
    warnings.warn(
        f"'{__name__}.log_dir' is deprecated.",
        DeprecationWarning, stacklevel=2
    )
    if sys.platform == "darwin":
        name = QCoreApplication.applicationName() or "Orange"
        logdir = os.path.join(os.path.expanduser("~/Library/Logs"), name)
    else:
        logdir = data_dir()

    try:
        os.makedirs(logdir, exist_ok=True)
    except OSError:
        pass
    return logdir


def widget_settings_dir(versioned=True):
    """
    Return the platform dependent directory where widgets save their settings.

    .. deprecated: 4.0.1
    """
    warnings.warn(
        f"'{__name__}.widget_settings_dir' is deprecated.",
        DeprecationWarning, stacklevel=2
    )
    from orangewidget.settings import widget_settings_dir
    return widget_settings_dir(versioned)


def widgets_entry_points():
    return Config.widgets_entry_points()


def splash_screen():
    return Config.splash_screen()


def application_icon():
    return Config.application_icon()



import unittest
from unittest.mock import Mock

from AnyQt.QtCore import QStringListModel, Qt, QSortFilterProxyModel
from AnyQt.QtTest import QTest
from AnyQt.QtWidgets import QLineEdit

from orangewidget.tests.base import GuiTest
from orangewidget.utils.itemmodels import PyListModel
from orangewidget.utils.listview import ListViewSearch, ListViewFilter


class TestListViewSearch(GuiTest):
    def setUp(self) -> None:
        super().setUp()
        self.lv = ListViewSearch()
        s = ["one", "two", "three", "four"]
        model = QStringListModel(s)
        self.lv.setModel(model)

    def tearDown(self) -> None:
        super().tearDown()
        self.lv.deleteLater()
        self.lv = None

    def test_list_view(self):
        num_items = 4
        self.assertEqual(num_items, self.lv.model().rowCount())

        filter_row = self.lv.findChild(QLineEdit)
        filter_row.grab()
        self.lv.grab()

        QTest.keyClick(filter_row, Qt.Key_E, delay=-1)
        self.assertListEqual(
            [False, True, False, True],
            [self.lv.isRowHidden(i) for i in range(num_items)],
        )
        QTest.keyClick(filter_row, Qt.Key_Backspace)
        self.assertListEqual(
            [False] * 4, [self.lv.isRowHidden(i) for i in range(num_items)]
        )
        QTest.keyClick(filter_row, Qt.Key_F)
        self.assertListEqual(
            [True, True, True, False],
            [self.lv.isRowHidden(i) for i in range(num_items)],
        )
        QTest.keyClick(filter_row, Qt.Key_Backspace)
        QTest.keyClick(filter_row, Qt.Key_T)
        self.assertListEqual(
            [True, False, False, True],
            [self.lv.isRowHidden(i) for i in range(num_items)],
        )
        QTest.keyClick(filter_row, Qt.Key_H)
        self.assertListEqual(
            [True, True, False, True],
            [self.lv.isRowHidden(i) for i in range(num_items)],
        )

    def test_insert_new_value(self):
        num_items = 4
        filter_row = self.lv.findChild(QLineEdit)
        filter_row.grab()
        self.lv.grab()

        QTest.keyClick(filter_row, Qt.Key_E, delay=-1)
        self.assertListEqual(
            [False, True, False, True],
            [self.lv.isRowHidden(i) for i in range(num_items)],
        )

        model = self.lv.model()
        if model.insertRow(model.rowCount()):
            index = model.index(model.rowCount() - 1, 0)
            model.setData(index, "six")

        self.assertListEqual(
            [False, True, False, True, True],
            [self.lv.isRowHidden(i) for i in range(num_items + 1)],
        )

    def test_empty(self):
        self.lv.setModel(QStringListModel([]))
        self.assertEqual(0, self.lv.model().rowCount())

        filter_row = self.lv.findChild(QLineEdit)
        filter_row.grab()
        self.lv.grab()

        QTest.keyClick(filter_row, Qt.Key_T)
        QTest.keyClick(filter_row, Qt.Key_Backspace)

    def test_PyListModel(self):
        model = PyListModel()
        view = ListViewSearch()
        view.setFilterString("two")
        view.setRowHidden = Mock(side_effect=view.setRowHidden)
        view.setModel(model)
        view.setRowHidden.assert_not_called()
        model.wrap(["one", "two", "three", "four"])
        view.setRowHidden.assert_called()
        self.assertTrue(view.isRowHidden(0))
        self.assertFalse(view.isRowHidden(1))
        self.assertTrue(view.isRowHidden(2))
        self.assertTrue(view.isRowHidden(3))


class TestListViewFilter(GuiTest):
    def test_filter(self):
        model = PyListModel()
        view = ListViewFilter()
        view._ListViewFilter__search.textEdited.emit("two")
        view.model().setSourceModel(model)
        model.wrap(["one", "two", "three", "four"])
        self.assertEqual(view.model().rowCount(), 1)
        self.assertEqual(model.rowCount(), 4)

    def test_set_model(self):
        view = ListViewFilter()
        self.assertRaises(Exception, view.setModel, PyListModel())

    def test_set_source_model(self):
        model = PyListModel()
        view = ListViewFilter()
        view.set_source_model(model)
        self.assertIs(view.model().sourceModel(), model)
        self.assertIs(view.source_model(), model)

    def test_set_proxy(self):
        proxy = QSortFilterProxyModel()
        view = ListViewFilter(proxy=proxy)
        self.assertIs(view.model(), proxy)

    def test_set_proxy_raises(self):
        self.assertRaises(Exception, ListViewFilter, proxy=PyListModel())

    def test_selection(self):
        view = ListViewFilter()
        view.model().setSourceModel(PyListModel(["one", "two", "three"]))
        sel_model = view.selectionModel()
        selected_row = 0
        index = view.model().index(selected_row, 0)
        sel_model.select(index, sel_model.ClearAndSelect)
        self.assertEqual(sel_model.selectedRows(0)[0].row(), selected_row)

        view._ListViewFilter__search.textEdited.emit("two")
        self.assertEqual(len(sel_model.selectedRows(0)), 0)

        view._ListViewFilter__search.textEdited.emit("")
        self.assertEqual(sel_model.selectedRows(0)[0].row(), selected_row)


if __name__ == "__main__":
    unittest.main()



from typing import Iterable, Optional
import warnings

from AnyQt.QtWidgets import QListView, QLineEdit, QStyle
from AnyQt.QtGui import QResizeEvent
from AnyQt.QtCore import (
    Qt,
    QAbstractItemModel,
    QModelIndex,
    QSortFilterProxyModel,
    QItemSelection,
    QSize,
    QItemSelectionModel,
    Signal,
)

from orangewidget.utils.itemmodels import signal_blocking


class ListViewFilter(QListView):
    """
    A QListView with implicit and transparent row filtering.
    """
    sigSelectionChanged = Signal(object)

    def __init__(
            self,
            *args,
            model: Optional[QAbstractItemModel] = None,
            proxy: Optional[QSortFilterProxyModel] = None,
            preferred_size: Optional[QSize] = None,
            **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__selection = QItemSelection()
        self.__search = QLineEdit(self, placeholderText="Filter...")
        self.__search.textEdited.connect(self.__on_text_edited)
        self.__preferred_size = preferred_size
        self.__layout()
        self.setMinimumHeight(100)

        if proxy is None:
            proxy = QSortFilterProxyModel(
                self, filterCaseSensitivity=Qt.CaseInsensitive
            )
        assert isinstance(proxy, QSortFilterProxyModel)
        super().setModel(proxy)
        self.set_source_model(model)
        self.selectionModel().selectionChanged.connect(self.__on_sel_changed)

    @property
    def selection(self) -> QItemSelection:
        return self.__selection

    def __on_sel_changed(
            self,
            selected: QItemSelection,
            deselected: QItemSelection
    ):
        selected = self.model().mapSelectionToSource(selected)
        deselected = self.model().mapSelectionToSource(deselected)
        if self.selectionMode() == QListView.SelectionMode.SingleSelection:
            self.__selection.clear()
        self.__selection.merge(selected, QItemSelectionModel.Select)
        self.__selection.merge(deselected, QItemSelectionModel.Deselect)
        self.sigSelectionChanged.emit(self.__selection)

    def __on_text_edited(self, string: str):
        with signal_blocking(self.selectionModel()):
            self.model().setFilterFixedString(string)
            self.__select()

    def __select(self):
        selection = self.model().mapSelectionFromSource(self.__selection)
        self.selectionModel().select(selection,
                                     QItemSelectionModel.ClearAndSelect)

    def setModel(self, _):
        raise TypeError("The model cannot be changed. "
                        "Use set_source_model() instead.")

    def set_source_model(self, model: QAbstractItemModel):
        self.model().setSourceModel(model)

    def source_model(self):
        return self.model().sourceModel()

    def updateGeometries(self):
        super().updateGeometries()
        self.__layout()

    def __layout(self):
        margins = self.viewportMargins()
        sh = self.__search.sizeHint()
        margins.setTop(sh.height())
        vscroll = self.verticalScrollBar()
        transient = self.style().styleHint(QStyle.SH_ScrollBar_Transient,
                                           None, vscroll)
        w = self.size().width()
        if vscroll.isVisibleTo(self) and not transient:
            w = w - vscroll.width() - 1
        self.__search.setGeometry(0, 0, w, sh.height())
        self.setViewportMargins(margins)

    def sizeHint(self) -> QSize:
        size = self.__preferred_size
        return size if size is not None else super().sizeHint()


class ListViewSearch(QListView):
    """
    An QListView with an implicit and transparent row filtering.
    """

    def __init__(self, *a, preferred_size=None, **ak):
        warnings.warn("ListViewSearch is deprecated and will be removed "
                      "in upcoming releases. Use ListViewFilter instead.",
                      DeprecationWarning)
        super().__init__(*a, **ak)
        self.__search = QLineEdit(self, placeholderText="Filter...")
        self.__search.textEdited.connect(self.__setFilterString)
        # Use an QSortFilterProxyModel for filtering. Note that this is
        # never set on the view, only its rows insertes/removed signals are
        # connected to observe an update row hidden state.
        self.__pmodel = QSortFilterProxyModel(
            self, filterCaseSensitivity=Qt.CaseInsensitive
        )
        self.__pmodel.rowsAboutToBeRemoved.connect(
            self.__filter_rowsAboutToBeRemoved
        )
        self.__pmodel.rowsInserted.connect(self.__filter_rowsInserted)
        self.__layout()
        self.preferred_size = preferred_size
        self.setMinimumHeight(100)

    def setFilterPlaceholderText(self, text: str):
        self.__search.setPlaceholderText(text)

    def filterPlaceholderText(self) -> str:
        return self.__search.placeholderText()

    def setFilterProxyModel(self, proxy: QSortFilterProxyModel) -> None:
        """
        Set an instance of QSortFilterProxyModel that will be used for filtering
        the model. The `proxy` must be a filtering proxy only; it MUST not sort
        the row of the model.
        The FilterListView takes ownership of the proxy.
        """
        self.__pmodel.rowsAboutToBeRemoved.disconnect(
            self.__filter_rowsAboutToBeRemoved
        )
        self.__pmodel.rowsInserted.disconnect(self.__filter_rowsInserted)
        self.__pmodel = proxy
        proxy.setParent(self)
        self.__pmodel.rowsAboutToBeRemoved.connect(
            self.__filter_rowsAboutToBeRemoved
        )
        self.__pmodel.rowsInserted.connect(self.__filter_rowsInserted)
        self.__pmodel.setSourceModel(self.model())
        self.__filter_reset()

    def filterProxyModel(self) -> QSortFilterProxyModel:
        return self.__pmodel

    def setModel(self, model: QAbstractItemModel) -> None:
        super().setModel(model)
        self.__pmodel.setSourceModel(model)
        self.__filter_reset()
        self.model().rowsInserted.connect(self.__model_rowInserted)
        self.model().modelReset.connect(self.__on_modelReset)

    def __on_modelReset(self):
        self.__filter_reset()
        self.__pmodel.setFilterFixedString("")
        self.__pmodel.setFilterFixedString(self.__search.text())

    def setRootIndex(self, index: QModelIndex) -> None:
        super().setRootIndex(index)
        self.__filter_reset()

    def __filter_reset(self):
        root = self.rootIndex()
        self.__filter(range(self.__pmodel.rowCount(root)))

    def __setFilterString(self, string: str):
        self.__pmodel.setFilterFixedString(string)

    def setFilterString(self, string: str):
        """Set the filter string."""
        self.__search.setText(string)
        self.__pmodel.setFilterFixedString(string)

    def filterString(self):
        """Return the filter string."""
        return self.__search.text()

    def __filter(self, rows: Iterable[int]) -> None:
        """Set hidden state for rows based on filter string"""
        root = self.rootIndex()
        pm = self.__pmodel
        for r in rows:
            self.setRowHidden(r, not pm.filterAcceptsRow(r, root))

    def __filter_set(self, rows: Iterable[int], state: bool):
        for r in rows:
            self.setRowHidden(r, state)

    def __filter_rowsAboutToBeRemoved(
        self, parent: QModelIndex, start: int, end: int
    ) -> None:
        fmodel = self.__pmodel
        mrange = QItemSelection(
            fmodel.index(start, 0, parent), fmodel.index(end, 0, parent)
        )
        mranges = fmodel.mapSelectionToSource(mrange)
        for mrange in mranges:
            self.__filter_set(range(mrange.top(), mrange.bottom() + 1), True)

    def __filter_rowsInserted(
        self, parent: QModelIndex, start: int, end: int
    ) -> None:
        fmodel = self.__pmodel
        mrange = QItemSelection(
            fmodel.index(start, 0, parent), fmodel.index(end, 0, parent)
        )
        mranges = fmodel.mapSelectionToSource(mrange)
        for mrange in mranges:
            self.__filter_set(range(mrange.top(), mrange.bottom() + 1), False)

    def __model_rowInserted(self, _, start: int, end: int) -> None:
        """
        Filter elements when inserted in list - proxy model's rowsAboutToBeRemoved
        is not called on elements that are hidden when inserting
        """
        self.__filter(range(start, end + 1))

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)

    def updateGeometries(self) -> None:
        super().updateGeometries()
        self.__layout()

    def __layout(self):
        margins = self.viewportMargins()
        search = self.__search
        sh = search.sizeHint()
        size = self.size()
        margins.setTop(sh.height())
        vscroll = self.verticalScrollBar()
        style = self.style()
        transient = style.styleHint(QStyle.SH_ScrollBar_Transient, None, vscroll)
        w = size.width()
        if vscroll.isVisibleTo(self) and not transient:
            w = w - vscroll.width() - 1
        search.setGeometry(0, 0, w, sh.height())
        self.setViewportMargins(margins)

    def sizeHint(self):
        return (
            self.preferred_size
            if self.preferred_size is not None
            else super().sizeHint()
        )


def main():
    from itertools import cycle
    from AnyQt.QtCore import QStringListModel
    from AnyQt.QtWidgets import QApplication, QWidget, QVBoxLayout

    app = QApplication([])
    w = QWidget()
    w.setLayout(QVBoxLayout())
    lv = ListViewFilter()
    lv.setUniformItemSizes(True)
    w.layout().addWidget(lv)
    c = cycle(list(map(chr, range(ord("A"), ord("Z")))))
    s = [f"{next(c)}{next(c)}{next(c)}{next(c)}" for _ in range(50000)]
    model = QStringListModel(s)
    lv.set_source_model(model)
    w.show()
    app.exec()


if __name__ == "__main__":
    main()



import enum
from datetime import date, datetime
from functools import partial
from itertools import filterfalse
from types import MappingProxyType as MappingProxy
from typing import (
    Sequence, Any, Mapping, Dict, TypeVar, Type, Optional, Container, Tuple,
)
from typing_extensions import Final

import numpy as np

from AnyQt.QtCore import (
    Qt, QObject, QAbstractItemModel, QModelIndex, QPersistentModelIndex, Slot,
    QLocale, QRect, QPointF, QSize, QLineF,
)
from AnyQt.QtGui import (
    QFont, QFontMetrics, QPalette, QColor, QBrush, QIcon, QPixmap, QImage,
    QPainter, QStaticText, QTransform, QPen
)
from AnyQt.QtWidgets import QStyledItemDelegate, QStyleOptionViewItem, \
    QApplication, QStyle

from orangewidget.utils.cache import LRUCache
from orangewidget.utils import enum_as_int, grapheme_slice

A = TypeVar("A")


def item_data(
        index: QModelIndex, roles: Sequence[int]
) -> Dict[int, Any]:
    """Query `index` for all `roles` and return them as a mapping"""
    model = index.model()
    datagetter = partial(model.data, index)
    values = map(datagetter, roles)
    return dict(zip(roles, values))


class ModelItemCache(QObject):
    """
    An item data cache for accessing QAbstractItemModel.data

    >>> cache = ModelItemCache()
    >>> cache.itemData(index, (Qt.DisplayRole, Qt.DecorationRole))
    {0: ...

    """
    #: The cache key is a tuple of the persistent index of the *parent*,
    #: row and column. The parent is used because of performance regression in
    #: Qt6 ~QPersistentModelIndex destructor when there are many (different)
    #: persistent indices registered with a model. Using parent, row, column
    #: coalesces these.
    #: NOTE: QPersistentModelIndex's hash changes when it is invalidated;
    #: it must be purged from __cache_data before that (see `__connect_helper`)
    __KEY = Tuple[QPersistentModelIndex, int, int]
    __slots__ = ("__model", "__cache_data")

    def __init__(self, *args, maxsize=100 * 200, **kwargs):
        super().__init__(*args, **kwargs)
        self.__model: Optional[QAbstractItemModel] = None
        self.__cache_data: 'LRUCache[ModelItemCache.__KEY, Any]' = LRUCache(maxsize)

    def __connect_helper(self, model: QAbstractItemModel) -> None:
        model.dataChanged.connect(self.invalidate)
        model.layoutAboutToBeChanged.connect(self.invalidate)
        model.modelAboutToBeReset.connect(self.invalidate)
        model.rowsAboutToBeInserted.connect(self.invalidate)
        model.rowsAboutToBeRemoved.connect(self.invalidate)
        model.rowsAboutToBeMoved.connect(self.invalidate)
        model.columnsAboutToBeInserted.connect(self.invalidate)
        model.columnsAboutToBeRemoved.connect(self.invalidate)
        model.columnsAboutToBeMoved.connect(self.invalidate)

    def __disconnect_helper(self, model: QAbstractItemModel) -> None:
        model.dataChanged.disconnect(self.invalidate)
        model.layoutAboutToBeChanged.disconnect(self.invalidate)
        model.modelAboutToBeReset.disconnect(self.invalidate)
        model.rowsAboutToBeInserted.disconnect(self.invalidate)
        model.rowsAboutToBeRemoved.disconnect(self.invalidate)
        model.rowsAboutToBeMoved.disconnect(self.invalidate)
        model.columnsAboutToBeInserted.disconnect(self.invalidate)
        model.columnsAboutToBeRemoved.disconnect(self.invalidate)
        model.columnsAboutToBeMoved.disconnect(self.invalidate)

    def setModel(self, model: QAbstractItemModel) -> None:
        if model is self.__model:
            return
        if self.__model is not None:
            self.__disconnect_helper(self.__model)
            self.__model = None
        self.__model = model
        self.__cache_data.clear()
        if model is not None:
            self.__connect_helper(model)

    def model(self) -> Optional[QAbstractItemModel]:
        return self.__model

    @Slot()
    def invalidate(self) -> None:
        """Invalidate all cached data."""
        self.__cache_data.clear()

    def itemData(
            self, index: QModelIndex, roles: Sequence[int]
    ) -> Mapping[int, Any]:
        """
        Return item data from `index` for `roles`.

        The returned mapping is a read only view of *all* data roles accessed
        for the index through this caching interface. It will contain at least
        data for `roles`, but can also contain other ones.
        """
        model = index.model()
        if model is not self.__model:
            self.setModel(model)
        key = QPersistentModelIndex(index.parent()), index.row(), index.column()
        try:
            item = self.__cache_data[key]
        except KeyError:
            data = item_data(index, roles)
            view = MappingProxy(data)
            self.__cache_data[key] = data, view
        else:
            data, view = item
            queryroles = tuple(filterfalse(data.__contains__, roles))
            if queryroles:
                data.update(item_data(index, queryroles))
        return view

    def data(self, index: QModelIndex, role: int) -> Any:
        """Return item data for `index` and `role`"""
        model = index.model()
        if model is not self.__model:
            self.setModel(model)
        key = QPersistentModelIndex(index.parent()), index.row(), index.column()
        try:
            item = self.__cache_data[key]
        except KeyError:
            data = item_data(index, (role,))
            view = MappingProxy(data)
            self.__cache_data[key] = data, view
        else:
            data, view = item
            if role not in data:
                data[role] = model.data(index, role)
        return data[role]


def cast_(type_: Type[A], value: Any) -> Optional[A]:
    # similar but not quite the same as qvariant_cast
    if value is None:
        return value
    if type(value) is type_:  # pylint: disable=unidiomatic-typecheck
        return value
    try:
        return type_(value)
    except Exception:  # pylint: disable=broad-except  # pragma: no cover
        return None


# QStyleOptionViewItem.Feature aliases as python int. Feature.__ior__
# implementation is slower then int.__ior__
_QStyleOptionViewItem_HasDisplay = enum_as_int(QStyleOptionViewItem.HasDisplay)
_QStyleOptionViewItem_HasCheckIndicator = enum_as_int(QStyleOptionViewItem.HasCheckIndicator)
_QStyleOptionViewItem_HasDecoration = enum_as_int(QStyleOptionViewItem.HasDecoration)


class _AlignmentFlagsCache(dict):
    # A cached int -> Qt.Alignment cache. Used to avoid temporary Qt.Alignment
    # flags object (de)allocation.
    def __missing__(self, key: int) -> Qt.AlignmentFlag:
        a = Qt.AlignmentFlag(key)
        self.setdefault(key, a)
        return a


_AlignmentCache: Mapping[int, Qt.Alignment] = _AlignmentFlagsCache()
_AlignmentMask = int(Qt.AlignHorizontal_Mask | Qt.AlignVertical_Mask)


def init_style_option(
        delegate: QStyledItemDelegate,
        option: QStyleOptionViewItem,
        index: QModelIndex,
        data: Mapping[int, Any],
        roles: Optional[Container[int]] = None,
) -> None:
    """
    Like `QStyledItemDelegate.initStyleOption` but fill in the fields from
    `data` mapping. If `roles` is not `None` init the `option` for the
    specified `roles` only.
    """
    # pylint: disable=too-many-branches
    option.styleObject = None
    option.index = index
    if roles is None:
        roles = data
    features = 0
    if Qt.DisplayRole in roles:
        value = data.get(Qt.DisplayRole)
        if value is not None:
            option.text = delegate.displayText(value, option.locale)
            features |= _QStyleOptionViewItem_HasDisplay
    if Qt.FontRole in roles:
        value = data.get(Qt.FontRole)
        font = cast_(QFont, value)
        if font is not None:
            font = font.resolve(option.font)
            option.font = font
            option.fontMetrics = QFontMetrics(option.font)
    if Qt.ForegroundRole in roles:
        value = data.get(Qt.ForegroundRole)
        foreground = cast_(QBrush, value)
        if foreground is not None:
            option.palette.setBrush(QPalette.Text, foreground)
    if Qt.BackgroundRole in roles:
        value = data.get(Qt.BackgroundRole)
        background = cast_(QBrush, value)
        if background is not None:
            option.backgroundBrush = background
    if Qt.TextAlignmentRole in roles:
        value = data.get(Qt.TextAlignmentRole)
        alignment = cast_(int, value)
        if alignment is not None:
            alignment = alignment & _AlignmentMask
            option.displayAlignment = _AlignmentCache[alignment]
    if Qt.CheckStateRole in roles:
        state = data.get(Qt.CheckStateRole)
        if state is not None:
            features |= _QStyleOptionViewItem_HasCheckIndicator
            state = cast_(int, state)
            if state is not None:
                option.checkState = state
    if Qt.DecorationRole in roles:
        value = data.get(Qt.DecorationRole)
        if value is not None:
            features |= _QStyleOptionViewItem_HasDecoration
        if isinstance(value, QIcon):
            option.icon = value
        elif isinstance(value, QColor):
            pix = QPixmap(option.decorationSize)
            pix.fill(value)
            option.icon = QIcon(pix)
        elif isinstance(value, QPixmap):
            option.icon = QIcon(value)
            option.decorationSize = (value.size() / value.devicePixelRatio()).toSize()
        elif isinstance(value, QImage):
            pix = QPixmap.fromImage(value)
            option.icon = QIcon(value)
            option.decorationSize = (pix.size() / pix.devicePixelRatio()).toSize()
    option.features |= QStyleOptionViewItem.ViewItemFeature(features)


class CachedDataItemDelegate(QStyledItemDelegate):
    """
    An QStyledItemDelegate with item model data caching.

    Parameters
    ----------
    roles: Sequence[int]
        A set of roles to query the model and fill the `QStyleOptionItemView`
        with. By specifying only a subset of the roles here the delegate can
        be speed up (e.g. if you know the model does not provide the relevant
        roles or you just want to ignore some of them).
    """
    __slots__ = ("roles", "__cache",)

    #: The default roles that are filled in initStyleOption
    DefaultRoles = (
        Qt.DisplayRole, Qt.TextAlignmentRole, Qt.FontRole, Qt.ForegroundRole,
        Qt.BackgroundRole, Qt.CheckStateRole, Qt.DecorationRole
    )

    def __init__(
            self, *args, roles: Sequence[int] = None, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        if roles is None:
            roles = self.DefaultRoles
        self.roles = tuple(roles)
        self.__cache = ModelItemCache(self)

    def cachedItemData(
            self, index: QModelIndex, roles: Sequence[int]
    ) -> Mapping[int, Any]:
        """
        Return a mapping of all roles for the index.

        .. note::
           The returned mapping contains at least `roles`, but will also
           contain all cached roles that were queried previously.
        """
        return self.__cache.itemData(index, roles)

    def cachedData(self, index: QModelIndex, role: int) -> Any:
        """Return the data for role from `index`."""
        return self.__cache.data(index, role)

    def initStyleOption(
            self, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        """
        Reimplemented.

        Use caching to query the model data. Also limit the roles queried
        from the model and filled in `option` to `self.roles`.
        """
        data = self.cachedItemData(index, self.roles)
        init_style_option(self, option, index, data, self.roles)


_Real = (float, np.floating)
_Integral = (int, np.integer)
_Number = _Integral + _Real
_String = (str, np.str_)
_DateTime = (date, datetime, np.datetime64)
_TypesAlignRight = _Number + _DateTime


class StyledItemDelegate(QStyledItemDelegate):
    """
    A `QStyledItemDelegate` subclass supporting a broader range of python
    and numpy types for display.

    E.g. supports `np.float*`, `np.(u)int`, `datetime.date`,
    `datetime.datetime`
    """
    #: Types that are displayed as real (decimal)
    RealTypes: Final[Tuple[type, ...]] = _Real
    #: Types that are displayed as integers
    IntegralTypes: Final[Tuple[type, ...]] = _Integral
    #: RealTypes and IntegralTypes combined
    NumberTypes: Final[Tuple[type, ...]] = _Number
    #: Date time types
    DateTimeTypes: Final[Tuple[type, ...]] = _DateTime

    def displayText(self, value: Any, locale: QLocale) -> str:
        """
        Reimplemented.
        """
        # NOTE: Maybe replace the if,elif with a dispatch a table
        if value is None:
            return ""
        elif type(value) is str:  # pylint: disable=unidiomatic-typecheck
            return value  # avoid copies
        elif isinstance(value, _Integral):
            return super().displayText(int(value), locale)
        elif isinstance(value, _Real):
            return super().displayText(float(value), locale)
        elif isinstance(value, _String):
            return str(value)
        elif isinstance(value, datetime):
            return value.isoformat(sep=" ")
        elif isinstance(value, date):
            return value.isoformat()
        elif isinstance(value, np.datetime64):
            return self.displayText(value.astype(datetime), locale)
        return super().displayText(value, locale)


_Qt_AlignRight = enum_as_int(Qt.AlignRight)
_Qt_AlignLeft = enum_as_int(Qt.AlignLeft)
_Qt_AlignHCenter = enum_as_int(Qt.AlignHCenter)
_Qt_AlignTop = enum_as_int(Qt.AlignTop)
_Qt_AlignBottom = enum_as_int(Qt.AlignBottom)
_Qt_AlignVCenter = enum_as_int(Qt.AlignVCenter)

_StaticTextKey = Tuple[str, QFont, Qt.TextElideMode, int]
_PenKey = Tuple[str, int]
_State_Mask = enum_as_int(
    QStyle.State_Selected | QStyle.State_Enabled | QStyle.State_Active
)


class DataDelegate(CachedDataItemDelegate, StyledItemDelegate):
    """
    A QStyledItemDelegate optimized for displaying fixed tabular data.

    This delegate will automatically display numeric and date/time values
    aligned to the right.

    Note
    ----
    Does not support text wrapping
    """
    __slots__ = (
        "__static_text_lru_cache", "__pen_lru_cache", "__style"
    )
    #: Types that are right aligned by default (when Qt.TextAlignmentRole
    #: is not defined by the model or is excluded from self.roles)
    TypesAlignRight: Final[Tuple[type, ...]] = _TypesAlignRight

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__static_text_lru_cache: LRUCache[_StaticTextKey, QStaticText]
        self.__static_text_lru_cache = LRUCache(100 * 200)
        self.__pen_lru_cache: LRUCache[_PenKey, QPen] = LRUCache(100)
        self.__style = None
        self.__max_text_length = 500

    def initStyleOption(
            self, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        data = self.cachedItemData(index, self.roles)
        init_style_option(self, option, index, data, self.roles)
        if data.get(Qt.TextAlignmentRole) is None \
                and Qt.TextAlignmentRole in self.roles \
                and isinstance(data.get(Qt.DisplayRole), _TypesAlignRight):
            option.displayAlignment = \
                (option.displayAlignment & ~Qt.AlignHorizontal_Mask) | \
                Qt.AlignRight

    def paint(
            self, painter: QPainter, option: QStyleOptionViewItem,
            index: QModelIndex
    ) -> None:
        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)
        widget = option.widget
        style = QApplication.style() if widget is None else widget.style()
        # Keep ref to style wrapper. This is ugly, wrong but the wrapping of
        # C++ QStyle instance takes ~5% unless the wrapper already exists.
        self.__style = style
        # Draw empty item cell
        opt_c = QStyleOptionViewItem(opt)
        opt_c.text = ""
        style.drawControl(QStyle.CE_ItemViewItem, opt_c, painter, widget)
        trect = style.subElementRect(QStyle.SE_ItemViewItemText, opt_c, widget)
        self.drawViewItemText(style, painter, opt, trect)

    def drawViewItemText(
            self, style: QStyle, painter: QPainter,
            option: QStyleOptionViewItem, rect: QRect
    ) -> None:
        """
        Draw view item text in `rect` using `style` and `painter`.
        """
        margin = style.pixelMetric(
            QStyle.PM_FocusFrameHMargin, None, option.widget) + 1
        rect = rect.adjusted(margin, 0, -margin, 0)
        font = option.font
        text = option.text
        st = self.__static_text_elided_cache(
            text, font, option.fontMetrics, option.textElideMode,
            rect.width()
        )
        tsize = st.size()
        textalign = enum_as_int(option.displayAlignment)
        text_pos_x = text_pos_y = 0.0

        if textalign & _Qt_AlignLeft:
            text_pos_x = rect.left()
        elif textalign & _Qt_AlignRight:
            text_pos_x = rect.x() + rect.width() - tsize.width()
        elif textalign & _Qt_AlignHCenter:
            text_pos_x = rect.x() + rect.width() / 2 - tsize.width() / 2

        if textalign & _Qt_AlignVCenter:
            text_pos_y = rect.y() + rect.height() / 2 - tsize.height() / 2
        elif textalign & _Qt_AlignTop:
            text_pos_y = rect.top()
        elif textalign & _Qt_AlignBottom:
            text_pos_y = rect.top() + rect.height() - tsize.height()

        painter.setPen(self.__pen_cache(option.palette, option.state))
        painter.setFont(font)
        painter.drawStaticText(QPointF(text_pos_x, text_pos_y), st)

    def __static_text_elided_cache(
            self, text: str, font: QFont, fontMetrics: QFontMetrics,
            elideMode: Qt.TextElideMode, width: int
    ) -> QStaticText:
        """
        Return a `QStaticText` instance for depicting the text with the `font`
        """
        try:
            return self.__static_text_lru_cache[text, font, elideMode, width]
        except KeyError:
            # limit text to some sensible length in case it is a whole epic
            # tale or similar. elidedText will parse all of it to glyphs which
            # can be slow.
            text_limited = self.__cut_text(text)
            st = QStaticText(fontMetrics.elidedText(text_limited, elideMode, width))
            st.prepare(QTransform(), font)
            # take a copy of the font for cache key
            key = text, QFont(font), elideMode, width
            self.__static_text_lru_cache[key] = st
            return st

    def __cut_text(self, text):
        if len(text) > self.__max_text_length:
            return grapheme_slice(text, end=self.__max_text_length)
        else:
            return text

    def __pen_cache(self, palette: QPalette, state: QStyle.State) -> QPen:
        """Return a QPen from the `palette` for `state`."""
        # NOTE: This method exists mostly to avoid QPen, QColor (de)allocations.
        key = palette.cacheKey(), enum_as_int(state) & _State_Mask
        try:
            return self.__pen_lru_cache[key]
        except KeyError:
            pen = QPen(text_color_for_state(palette, state))
            self.__pen_lru_cache[key] = pen
            return pen


def text_color_for_state(palette: QPalette, state: QStyle.State) -> QColor:
    """Return the appropriate `palette` text color for the `state`."""
    cgroup = QPalette.Normal if state & QStyle.State_Active else QPalette.Inactive
    cgroup = cgroup if state & QStyle.State_Enabled else QPalette.Disabled
    role = QPalette.HighlightedText if state & QStyle.State_Selected else QPalette.Text
    return palette.color(cgroup, role)


class BarItemDataDelegate(DataDelegate):
    """
    An delegate drawing a horizontal bar below its text.

    Can be used to visualise numerical column distribution.

    Parameters
    ----------
    parent: Optional[QObject]
        Parent object
    color: QColor
        The default color for the bar. If not set then the palette's
        foreground role is used.
    penWidth: int
        The bar pen width.
    barFillRatioRole: int
        The item model role used to query the bar fill ratio (see
        :method:`barFillRatioData`)
    barColorRole: int
        The item model role used to query the bar color.
    """
    __slots__ = (
        "color", "penWidth", "barFillRatioRole", "barColorRole",
        "__line", "__pen"
    )

    def __init__(
            self, parent: Optional[QObject] = None, color=QColor(), penWidth=5,
            barFillRatioRole=Qt.UserRole + 1, barColorRole=Qt.UserRole + 2,
            **kwargs
    ):
        super().__init__(parent, **kwargs)
        self.color = color
        self.penWidth = penWidth
        self.barFillRatioRole = barFillRatioRole
        self.barColorRole = barColorRole
        # Line and pen instances reused
        self.__line = QLineF()
        self.__pen = QPen(color, penWidth, Qt.SolidLine, Qt.RoundCap)

    def barFillRatioData(self, index: QModelIndex) -> Optional[float]:
        """
        Return a number between 0.0 and 1.0 indicating the bar fill ratio.

        The default implementation queries the model for `barFillRatioRole`
        """
        return cast_(float, self.cachedData(index, self.barFillRatioRole))

    def barColorData(self, index: QModelIndex) -> Optional[QColor]:
        """
        Return the color for the bar.

        The default implementation queries the model for `barColorRole`
        """
        return cast_(QColor, self.cachedData(index, self.barColorRole))

    def sizeHint(
            self, option: QStyleOptionViewItem, index: QModelIndex
    ) -> QSize:
        sh = super().sizeHint(option, index)
        pw, vmargin = self.penWidth, 1
        sh.setHeight(sh.height() + pw + vmargin)
        return sh

    def paint(
            self, painter: QPainter, option: QStyleOptionViewItem,
            index: QModelIndex
    ) -> None:
        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)
        widget = option.widget
        style = QApplication.style() if widget is None else widget.style()
        self.__style = style
        text = opt.text
        opt.text = ""
        style.drawControl(QStyle.CE_ItemViewItem, opt, painter, widget)

        textrect = style.subElementRect(
            QStyle.SE_ItemViewItemText, opt, widget)

        ratio = self.barFillRatioData(index)
        if ratio is not None and 0. <= ratio <= 1.:
            color = self.barColorData(index)
            if color is None:
                color = self.color
            if not color.isValid():
                color = opt.palette.color(QPalette.WindowText)
            rect = option.rect
            pw = self.penWidth
            hmargin = 3 + pw / 2  # + half pen width for the round line cap
            vmargin = 1
            textoffset = pw + vmargin * 2
            baseline = rect.bottom() - textoffset / 2
            width = (rect.width() - 2 * hmargin) * ratio
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing)
            pen = self.__pen
            pen.setColor(color)
            pen.setWidth(pw)
            painter.setPen(pen)
            line = self.__line
            left = rect.left() + hmargin
            line.setLine(left, baseline, left + width, baseline)
            painter.drawLine(line)
            painter.restore()
            textrect.adjust(0, 0, 0, -textoffset)

        opt.text = text
        self.drawViewItemText(style, painter, opt, textrect)
